# Copyright 2015 ZTE, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from networking_zte.ml2 import driver_api as api
from neutron.plugins.ml2 import driver_context as context


class SecurityGroupContext(
        context.MechanismDriverContext, api.SecurityGroupContext):

    def __init__(
            self, plugin, plugin_context,
            security_group, original_security_group=None):
        super(SecurityGroupContext, self).__init__(plugin, plugin_context)
        self._security_group = security_group
        self._original_security_group = original_security_group

    @property
    def current(self):
        return self._security_group

    @property
    def original(self):
        return self._original_security_group


class SecurityGroupRuleContext(
        context.MechanismDriverContext, api.SecurityGroupRuleContext):

    def __init__(
            self, plugin, plugin_context,
            security_group_rule, original_security_group_rule=None):
        super(SecurityGroupRuleContext, self).__init__(plugin, plugin_context)
        self._security_group_rule = security_group_rule
        self._original_security_group_rule = original_security_group_rule

    @property
    def current(self):
        return self._security_group_rule

    @property
    def original(self):
        return self._original_security_group_rule


class QosPolicyContext(
        context.MechanismDriverContext, api.QosPolicyContext):

    def __init__(
            self, plugin, plugin_context,
            qos_policy, original_qos_policy=None):
        super(QosPolicyContext, self).__init__(plugin, plugin_context)
        self._qos_policy = qos_policy
        self._original_qos_policy = original_qos_policy

    @property
    def current(self):
        return self._qos_policy

    @property
    def original(self):
        return self._original_qos_policy


class QosBandWidthRuleContext(
        context.MechanismDriverContext, api.QosBandWidthRuleContext):

    def __init__(
            self, plugin, plugin_context,
            qos_bandwidth_rule, original_qos_bandwidth_rule=None):
        super(QosBandWidthRuleContext, self).__init__(plugin, plugin_context)
        self._qos_bandwidth_rule = qos_bandwidth_rule
        self._original_qos_bandwidth_rule = original_qos_bandwidth_rule

    @property
    def current(self):
        return self._qos_bandwidth_rule

    @property
    def original(self):
        return self._original_qos_bandwidth_rule



# Copyright 2017 ZTE, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


from networking_zte.common import servermanager
from networking_zte.utils import cmcc_util
try:
    from neutron.agent.common import config as config
except Exception:
    from neutron.conf.agent import common as config
try:
    from neutron.agent.l3 import config as l3_config
    from neutron.agent.l3 import ha
except Exception:
    from neutron.conf.agent.l3 import config as l3_config
    from neutron.conf.agent.l3 import ha as ha_conf
from neutron.agent.linux import external_process
from neutron.agent.linux import interface
from neutron.agent.linux import iptables_manager
try:
    from neutron.agent.metadata import config as metadata_config
except Exception:
    from neutron.conf.agent.metadata import config as metadata_config
from neutron.agent import rpc as agent_rpc
from neutron.common import config as common_config
try:
    from neutron_lib import constants as l3_constants
except Exception:
    from neutron.common import constants as l3_constants
from neutron.common import eventlet_utils
from neutron.common import rpc as n_rpc
from neutron.common import topics
try:
    from neutron_lib.utils import net as utils
except Exception:
    from neutron.common import utils

try:
    from neutron import context as context
except Exception:
    from neutron_lib import context as context
from neutron import manager
try:
    from neutron.openstack.common import loopingcall
    from neutron.openstack.common import service
except Exception:
    from neutron.agent.linux import pd
    from neutron.agent.linux import ra
    from oslo_service import loopingcall
    from oslo_service import service
try:
    from neutron import context as qcontext
except Exception:
    from neutron_lib import context as qcontext
try:
    from neutron.db.models import l3 as l3_db
except Exception:
    from neutron.db import l3_db
from neutron import service as neutron_service
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging
import sys
import time


eventlet_utils.monkey_patch()


# NS_PREFIX = namespaces.NS_PREFIX
# INTERNAL_DEV_PREFIX = namespaces.INTERNAL_DEV_PREFIX
# EXTERNAL_DEV_PREFIX = namespaces.EXTERNAL_DEV_PREFIX
# LOG = LOG.getLogger(__name__)
# NS_PREFIX = 'qrouter-'
# INTERNAL_DEV_PREFIX = 'qr-'
# EXTERNAL_DEV_PREFIX = 'qg-'
# RPC_LOOP_INTERVAL = 1
# FLOATING_IP_CIDR_SUFFIX = '/32'
#
# SUBNET  = 'subnet'
# PORT    = 'port'
# ROUTER  = 'router'
# FLOATING_IP = 'floating-ip'
# VXLAN_TUNNEL = 'vxlan-tunnel'
# SECURITY_GROUP = 'sg'
# SECURITY_GROUP_RULE = 'sg-rule'
# CLASSIFIER = 'classifier'

SUCCESS_CODES = range(200, 207)
FAILURE_CODES = [400, 401, 404, 409]

# Number of routers to fetch from server at a time on resync.
# Needed to reduce load on server side and to speed up resync on agent side.
SYNC_ROUTERS_MAX_CHUNK_SIZE = 16

rest_proxy_opts = [
    cfg.ListOpt('servers', default=['localhost:8800'],
                help=_("A comma separated list of Big Switch or Floodlight "
                       "servers and port numbers. The plugin proxies the "
                       "requests to the Big Switch/Floodlight server, "
                       "which performs the networking configuration. Only one"
                       "server is needed per deployment, but you may wish to"
                       "deploy multiple servers to support failover.")),
    cfg.StrOpt('server_auth', default=None, secret=True,
               help=_("The username and password for authenticating against "
                      " the Big Switch or Floodlight controller.")),
    cfg.StrOpt('base_url', default='/restconf/operations/zenic-vdcapp-model:',
               help=_("Base URL for restconf server ")),
    cfg.BoolOpt('server_ssl', default=False,
                help=_("If True, Use SSL when connecting to the Big Switch or "
                       "Floodlight controller.")),
    cfg.BoolOpt('ssl_sticky', default=True,
                help=_("Trust and store the first certificate received for "
                       "each controller address and use it to validate future "
                       "connections to that address.")),
    cfg.BoolOpt('no_ssl_validation', default=False,
                help=_("Disables SSL certificate validation for controllers")),
    cfg.BoolOpt('cache_connections', default=True,
                help=_("Re-use HTTP/HTTPS connections to the controller.")),
    cfg.StrOpt('ssl_cert_directory',
               default='/etc/neutron/plugins/proxyagent/znic/ssl',
               help=_("Directory containing ca_certs and host_certs "
                      "certificate directories.")),
    cfg.BoolOpt('sync_data', default=False,
                help=_("Sync data on connect")),
    cfg.BoolOpt('auto_sync_on_failure', default=True,
                help=_("If neutron fails to create a resource because "
                       "the backend controller doesn't know of a dependency, "
                       "the plugin automatically triggers a full data "
                       "synchronization to the controller.")),
    cfg.IntOpt('consistency_interval', default=60,
               help=_("Time between verifications that the backend controller "
                      "database is consistent with Neutron")),
    cfg.IntOpt('server_timeout', default=10,
               help=_("Maximum number of seconds to wait for proxy request "
                      "to connect and complete.")),
    cfg.IntOpt('thread_pool_size', default=4,
               help=_("Maximum number of threads to spawn to handle large "
                      "volumes of port creations.")),
    cfg.StrOpt('neutron_id', default='neutron-' + utils.get_hostname(),
               deprecated_name='quantum_id',
               help=_("User defined identifier for this Neutron deployment")),
    cfg.BoolOpt('add_meta_server_route', default=True,
               help=_("Flag to decide if a route to the metadata server "
                      "should be injected into the VM")),
    cfg.StrOpt('zenic_version', default="50.1",
               help=_("Version number of the zenic controller corresponding "
                      "to the zenic agent")),
    cfg.StrOpt('up_gw_status', default="True",
               help=_("up gwport status bind to router")),
    cfg.StrOpt('op_version', default="",
               help=_("Version number of the zenic plugins")),
    cfg.StrOpt('sync_router', default="False",
               help=_("Flag to decide if plugin use sync router function")),
]

LOG = logging.getLogger(__name__)


class ZenicPluginApi(agent_rpc.PluginApi):
    pass


class ZenicL3RestServerPool(servermanager.ServerPool):
    """Zenic L3 Rest Server Pool for Zenic L3 Agent.

    This server pool has the router and floatingip operations
    of create, update and delete, to the Zenic Controller.
    """

    def __init__(
            self,
            servers,
            auth,
            zenic_version,
            ssl,
            no_ssl_validation,
            ssl_sticky,
            ssl_cert_directory,
            consistency_interval,
            timeout=False,
            cache_connections=True,
            base_uri='/restconf/operations/zenic-vdcapp-model:',
            op_version="",
            success_codes=SUCCESS_CODES,
            failure_codes=FAILURE_CODES,
            name='ZenicL3RestProxy'):
        super(
            ZenicL3RestServerPool,
            self).__init__(
            servers,
            auth,
            ssl,
            no_ssl_validation,
            ssl_sticky,
            ssl_cert_directory,
            consistency_interval,
            timeout,
            cache_connections,
            base_uri,
            success_codes,
            failure_codes,
            name)
        version = zenic_version.split('.')
        version = version[0] + version[1]
        LOG.debug(_("zenic_version = %s"), version)
        if (version.isdigit() is False) or (int(version) < 403):
            LOG.error(_("zenic_version error!zenic_version = %s"), version)
        self.zenic_version = int(version)
        self.floating_status = {}
        self.op_version = op_version
        self.cmcc = cmcc_util.CmccUtil()

    @staticmethod
    def validate_dict(instance, key, default_val):
        return instance[key] if (key in instance and
                                 instance[key]) else default_val

    def construct_router_rest_msg(self, router_info, action):
        # LOG.info(_('construct_router_info r_info:{0}'.format(router_info)))
        if action == 'DELETE' or action == 'GET':
            router_rest_data = {"input": {"id": router_info}}
            return router_rest_data
        else:
            internal_interfaces = router_info.get('_interfaces', set())
            internal_interfaces_subnets = list()
            if internal_interfaces:
                internal_interfaces_subnets =\
                    [intf['subnets'] for intf in internal_interfaces]
                LOG.debug("subnets:%s", internal_interfaces_subnets)
                for subnet in internal_interfaces_subnets:
                    LOG.debug('subnet:%s', subnet)
            routes_injects = router_info.get('routes', set())
            LOG.debug('internal_interfaces:{0}'.format(internal_interfaces))
            real_gw_port = router_info.get('gw_port_id', '')
            if not real_gw_port:
                real_gw_port = ''
            real_gw_port = self.cmcc.filter_gw_port(
                self.op_version, router=router_info, gw_port=real_gw_port)
            d = []
            for c in internal_interfaces_subnets:
                a = [subnet['id'] for subnet in c]
                for i in a:
                    d.append(i)
            b = [router_inject for router_inject in routes_injects]
            router_rest_data =\
                {"input": {"id": router_info['id'],
                           "name": router_info['name'],
                           "admin_state_up": router_info['admin_state_up'],
                           "tenant_id": router_info['tenant_id'],
                           "ext-gw-port": real_gw_port,
                           "enable_snat": router_info.get('enable_snat', True),
                           "router-interfaces": d,
                           "routes": b}
                 }

            if self.zenic_version > 403:
                if '_floatingips' in router_info:
                    routes_floatingips = router_info.get('_floatingips', set())
                    floating_ip = []
                    for routes_floatingip in routes_floatingips:
                        if 'host' in routes_floatingip:
                            del routes_floatingip['host']
                        if 'description' in routes_floatingip:
                            del routes_floatingip['description']
                        if 'dns_name' in routes_floatingip:
                            del routes_floatingip['dns_name']
                        if 'created_at' in routes_floatingip:
                            del routes_floatingip['created_at']
                        if 'updated_at' in routes_floatingip:
                            del routes_floatingip['updated_at']
                        if 'dns_domain' in routes_floatingip:
                            del routes_floatingip['dns_domain']
                        if 'revision_number' in routes_floatingip:
                            del routes_floatingip['revision_number']
                        if 'fixed_ip_address_scope' in routes_floatingip:
                            del routes_floatingip['fixed_ip_address_scope']
                        if 'project_id' in routes_floatingip:
                            del routes_floatingip['project_id']
                        if 'floating_port_id' in routes_floatingip:
                            del routes_floatingip['floating_port_id']
                        routes_floatingip['status'] = \
                            l3_constants.FLOATINGIP_STATUS_ACTIVE
                        floating_ip.append(routes_floatingip)
                        if router_info['id'] not in self.floating_status:
                            self.floating_status[router_info['id']] = {
                                routes_floatingip['id']:
                                l3_constants.FLOATINGIP_STATUS_ACTIVE}
                        else:
                            status = self.floating_status[router_info['id']]
                            status[routes_floatingip['id']] = \
                                l3_constants.FLOATINGIP_STATUS_ACTIVE
                    input = router_rest_data['input']
                    input["floating-ips"] = floating_ip
                else:
                    input = router_rest_data['input']
                    input["floating-ips"] = []
        rest_info = '\n'
        for k, v in router_rest_data['input'].items():
            rest_info += '{0} is {1}\n'.format(k, v)
            rest_info += '{0} is {1}\n'.format(k, v)
        LOG.debug('rest_router_data:{0}'.format(rest_info))
        return router_rest_data

    def rest_update_router(self, router_info):
        data = self.construct_router_rest_msg(router_info, 'ADD')
        resource = 'update-router'
        err_str = "Unable to create remote router: %s"
        resp = self.rest_action('POST', resource, data, err_str)
        if resp[0] == 0:
            return False
        return True

    def rest_delete_router(self, router_id):
        LOG.debug('While deleting,'
                  'construct_router_info router_id:{0}'.format(router_id))
        data = self.construct_router_rest_msg(router_id, 'DELETE')
        resource = 'del-router'
        err_str = "Unable to delete remote router: %s"
        resp = self.rest_action('POST', resource, data, err_str)
        if resp[0] == 0:
            return False
        return True

    @staticmethod
    def construct_all_routers_rest_msg(router_info):
        router_rest_data = {
            "input": {
                "router-list": list(router_info)}
        }
        return router_rest_data

    def rest_all_router_ids(self, all_router_ids):
        LOG.info(_('While rest_all_router_ids, '
                   'construct_router_info %s'), all_router_ids)
        data = self.construct_all_routers_rest_msg(all_router_ids)
        resource = 'sync-router-info'
        err_str = "Unable to rest_all_router_ids: %s"
        self.rest_action('POST', resource, data, err_str)
        return True


class L3PluginApi(object):
    """Agent side of the l3 agent RPC API.

    API version history:
        1.0 - Initial version.
        1.1 - Floating IP operational status updates

    """

    BASE_RPC_API_VERSION = '1.0'

    def __init__(self, topic, host):
        self.host = host
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)

    def get_routers(self, context, router_ids=None):
        """Make a remote process call to retrieve the sync data for routers."""
        cctxt = self.client.prepare()
        return cctxt.call(context, 'sync_routers', host=self.host,
                          router_ids=router_ids)

    def get_service_plugin_list(self, context):
        """Make a call to get the list of activated services."""
        cctxt = self.client.prepare(version='1.3')
        return cctxt.call(context, 'get_service_plugin_list')

    def update_floatingip_statuses(self, context, router_id, fip_statuses):
        """Call the plugin update floating IPs's operational status."""
        cctxt = self.client.prepare(version='1.1')
        return cctxt.call(context, 'update_floatingip_statuses',
                          router_id=router_id, fip_statuses=fip_statuses)


class RouterInfo(object):

    def __init__(self, router_id, root_helper, router):
        self.router_id = router_id
        self.ex_gw_port = None
        self._snat_enabled = None
        self._snat_action = None
        self.internal_ports = []
        self.root_helper = root_helper
        # Invoke the setter for establishing initial SNAT action
        self.router = router
        self.ns_name = None
        self.iptables_manager = iptables_manager.IptablesManager(
            root_helper=root_helper,
            # FIXME(danwent): use_ipv6=True,
            namespace=self.ns_name)
        self.routes = []

    @property
    def router(self):
        return self._router

    @router.setter
    def router(self, value):
        self._router = value
        if not self._router:
            return
        # enable_snat by default if it wasn't specified by plugin
        self._snat_enabled = self._router.get('enable_snat', True)
        # Set a SNAT action for the router
        if self._router.get('gw_port'):
            self._snat_action = ('add_rules' if self._snat_enabled
                                 else 'remove_rules')
        elif self.ex_gw_port:
            # Gateway port was removed, remove rules
            self._snat_action = 'remove_rules'


def get_rest_proxy_conf():
    conf_proxy = cfg.ConfigOpts()
    conf_proxy.register_opts(rest_proxy_opts, 'RESTPROXY')
    conf_proxy(args=[], default_config_files=['/etc/neutron/plugin.ini'])
    return conf_proxy


def register_opts(conf):
    conf.register_opts(l3_config.OPTS)
    conf.register_opts(metadata_config.SHARED_OPTS)
    try:
        conf.register_opts(ha.OPTS)
    except Exception:
        ha_conf.register_l3_agent_ha_opts(conf)
    config.register_interface_driver_opts_helper(conf)
    config.register_agent_state_opts_helper(conf)
    conf.register_opts(interface.OPTS)
    conf.register_opts(external_process.OPTS)
    try:
        config.register_use_namespaces_opts_helper(conf)
    except Exception:
        conf.register_opts(pd.OPTS)
        conf.register_opts(ra.OPTS)
        config.register_availability_zone_opts_helper(conf)


class ZenicAgent(manager.Manager):
    """Manager for L3NatAgent

        API version history:
        1.0 initial Version
        1.1 changed the type of the routers parameter
            to the routers_updated method.
            It was previously a list of routers in dict format.
            It is now a list of router IDs only.
            Per rpc versioning rules,  it is backwards compatible.
    """
    RPC_API_VERSION = '1.1'

    OPTS = [
        cfg.StrOpt('external_network_bridge', default='br-ex',
                   help=_("Name of bridge used for external network "
                          "traffic.")),
        cfg.IntOpt('metadata_port',
                   default=9697,
                   help=_("TCP Port used by Neutron metadata namespace "
                          "proxy.")),
        cfg.IntOpt('send_arp_for_ha',
                   default=0,
                   help=_("Send this many gratuitous ARPs for HA setup, if "
                          "less than or equal to 0, the feature is disabled")),
        cfg.StrOpt('router_id', default='',
                   help=_("If namespaces is disabled, the l3 agent can only"
                          " configure a router that has the matching router "
                          "ID.")),
        cfg.BoolOpt('handle_internal_only_routers',
                    default=True,
                    help=_("Agent should implement routers with no gateway")),
        cfg.StrOpt('gateway_external_network_id', default='',
                   help=_("UUID of external network for routers implemented "
                          "by the agents.")),
        cfg.BoolOpt('enable_metadata_proxy', default=True,
                    help=_("Allow running metadata proxy.")),
        cfg.BoolOpt('router_delete_namespaces', default=False,
                    help=_("Delete namespace after removing a router.")),
        cfg.StrOpt('metadata_proxy_socket',
                   default='$state_path/metadata_proxy',
                   help=_('Location of Metadata Proxy UNIX domain '
                          'socket')),
    ]
    target = oslo_messaging.Target(version='1.2')

    def __init__(self, host, conf=None):
        if conf:
            self.conf = conf
        else:
            self.conf = cfg.CONF
        self.root_helper = config.get_root_helper(self.conf)
        self.router_info = {}
        self.context = context.get_admin_context_without_session()
        self.plugin_rpc = L3PluginApi(topics.L3PLUGIN, host)
        self.fullsync = True
        self.updated_routers = set()
        self.removed_routers = set()
        self.sync_progress = False
        self.servers = None
        self.fail_update_rest_router_id = set()
        self.fail_delete_rest_router_id = set()
        conf_proxy = get_rest_proxy_conf()
        self.whole_syncing_not_succ_yet = False
        self.sync_protect_tick = 0
        self.agent_id = 'zenic-agent-%s' % self.conf.host
        self.zenic_rpc = ZenicPluginApi(topics.PLUGIN)
        self.up_gw_status = conf_proxy.RESTPROXY.up_gw_status
        self.sync_router = conf_proxy.RESTPROXY.sync_router
        if conf_proxy.RESTPROXY.servers is not '':
            self.servers = ZenicL3RestServerPool(
                conf_proxy.RESTPROXY.servers,
                conf_proxy.RESTPROXY.server_auth,
                conf_proxy.RESTPROXY.zenic_version,
                conf_proxy.RESTPROXY.server_ssl,
                conf_proxy.RESTPROXY.no_ssl_validation,
                conf_proxy.RESTPROXY.ssl_sticky,
                conf_proxy.RESTPROXY.ssl_cert_directory,
                conf_proxy.RESTPROXY.consistency_interval,
                conf_proxy.RESTPROXY.server_timeout,
                conf_proxy.RESTPROXY.cache_connections,
                conf_proxy.RESTPROXY.base_url,
                conf_proxy.RESTPROXY.op_version)
        # self.rpc_loop = loopingcall.FixedIntervalLoopingCall(
        #     self._rpc_loop)
        # self.rpc_loop.start(interval=RPC_LOOP_INTERVAL)
        super(ZenicAgent, self).__init__(host=host)
        self.target_ex_net_id = None
        self.sync_routers_chunk_size = SYNC_ROUTERS_MAX_CHUNK_SIZE

    def router_deleted(self, context, router_id):
        """Deal with router deletion RPC message."""
        LOG.info(_('zenic_agent Got router deleted notification for %s'),
                 router_id)
        try:
            with lockutils.lock("zenic_agent_remove_router"):
                LOG.debug('router_deleted Got remove mutex')
                self.removed_routers.add(router_id)
        except Exception as e:
            LOG.debug("lockutils, except:%s", str(e))

    def routers_updated(self, context, routers):
        """Deal with routers modification and creation RPC message."""
        LOG.info(_(' zenic_agent Got routers updated notification :%s'),
                 routers)
        if routers:
            # This is needed for backward compatibility
            if isinstance(routers[0], dict):
                routers = [router['id'] for router in routers]
            try:
                with lockutils.lock("zenic_agent_update_router"):
                    LOG.info(_('routers_updated Got update mutex'))
                    self.updated_routers.update(routers)
            except Exception as e:
                LOG.debug("lockutils, except:%s", str(e))

    def router_added_to_agent(self, context, payload):
        pass

    def _process_update_rest_fail_routers(self, routers):
        for r in routers:
            if self.servers:
                if self.servers.rest_update_router(r):
                    self.fail_update_rest_router_id.discard(r['id'])

    def _sync_all_valid_routers(self, routers):
        # pool = eventlet.GreenPool()
        return self.servers.rest_all_router_ids(routers)

    def _process_routers(self, routers):
        # pool = eventlet.GreenPool()
        for r in routers:
            if self.servers:
                if not self.servers.rest_update_router(r):
                    self.fail_update_rest_router_id.add(r['id'])
                else:
                    internal_interfaces = r.get('_interfaces', set())
                    internal_interfaces_devices = list()
                    if internal_interfaces:
                        internal_interfaces_devices = \
                            [intf['mac_address']
                             for intf in internal_interfaces]
                    for device in internal_interfaces_devices:
                        LOG.info(_('up router internal %s status'), device)
                        self.zenic_rpc.update_device_up(
                            self.context,
                            device,
                            self.agent_id,
                            self.conf.host)
                    if self.up_gw_status == "True":
                        if r.get('gw_port_id'):
                            device = r.get('gw_port')['mac_address']
                            LOG.info(_('up router gw %s status'), device)
                            self.zenic_rpc.update_device_up(
                                self.context,
                                device,
                                self.agent_id,
                                self.conf.host)
                    if r['id'] in self.servers.floating_status:
                        LOG.debug(_('update floatingip_statuses: %s'),
                                  self.servers.floating_status[r['id']])
                        self.plugin_rpc.update_floatingip_statuses(
                            self.context,
                            r['id'],
                            self.servers.floating_status[r['id']])
                        self.servers.floating_status[r['id']] = {}
                        if '_floatingips' not in r:
                            del self.servers.floating_status[r['id']]

    @staticmethod
    def get_all_router_ids(context):
        query = context.session.query(l3_db.Router.id)
        return [item[0] for item in query]

    def _sync_all_router_restconf(self, context):
        admin_context = qcontext.get_admin_context()
        routers = self.get_all_router_ids(admin_context)
        LOG.info(_('zenic_agent, starting, get_all_routers:%s'), routers)
        routers = set(router for router in routers)
        self.updated_routers.update(routers)
        now_valid_routers = routers & self.fail_delete_rest_router_id
        self.fail_delete_rest_router_id =\
            self.fail_delete_rest_router_id - now_valid_routers
        if self._sync_all_valid_routers(routers):
            LOG.debug('zenic_agent,  now_routers:{0}'.format(routers))
            self.whole_syncing_not_succ_yet = True

    def _process_routers_loop(self):
        LOG.debug("Starting _process_routers_loop")
        if not self.whole_syncing_not_succ_yet:
            if self.sync_router == "True":
                self._sync_all_router_restconf(self.context)
        while True:
            self._process_router_update()

    def after_start(self):
        LOG.info(_("zenic agent started"))
        self._process_routers_loop()
        # When L3 agent is ready, we immediately do a full sync
        # self.periodic_sync_routers_task(self.context)

    # @lockutils.synchronized('zenic-agent', 'neutron-')
    def _process_router_update(self):
        # _rpc_loop and _sync_routers_task will not be
        # executed in the same time because of lock.
        # so we can clear the value of updated_routers
        # and removed_routers
        self.sync_protect_tick += 1
        time.sleep(0.01)
        try:
            if self.updated_routers:
                LOG.info(_("zenic Starting RPC loop for %d updated routers"),
                         len(self.updated_routers))
                try:
                    with lockutils.lock("zenic_agent_update_router"):
                        LOG.info(_('_process_router_update Got update mutex'))
                        now_processed = list(self.updated_routers)
                        router_ids = list(self.updated_routers)
                        self.updated_routers.clear()
                except Exception as e:
                    LOG.error(_("lockutils, except:%s"), str(e))

                # fetch routers by chunks to reduce the load on server and to
                # start router processing earlier
                for i in range(
                        0,
                        len(router_ids),
                        self.sync_routers_chunk_size):
                    try:
                        now_processed_feched = now_processed[i:i +
                            self.sync_routers_chunk_size]
                        time_start_get_router = time.time()
                        routers = self.plugin_rpc.get_routers(
                            self.context,
                            router_ids[i:i + self.sync_routers_chunk_size])
                    except oslo_messaging.MessagingTimeout:
                        try:
                            time.sleep(5)
                            routers = self.plugin_rpc.get_routers(
                                self.context,
                                router_ids[i:i + self.sync_routers_chunk_size])
                        except oslo_messaging.MessagingTimeout:
                            LOG.error(_(
                                "Second time out error!!!,router_ids:%s"),
                                router_ids)
                            continue

                    LOG.debug('Processing routers: %r', routers)
                    fetched = set([r['id'] for r in routers])
                    try:
                        with lockutils.lock("zenic_agent_remove_router"):
                            self.removed_routers.update(
                                set(now_processed_feched) - fetched)
                    except Exception as e:
                        LOG.error(_("lockutils, except:%s"), str(e))

                    time_start_proc_router = time.time()
                    self._process_routers(routers)
                    LOG.info(_(
                        "Zenic ended _process_routers %(num)d routers, "
                        "get time: %(time)s, proc time: %(time2)s"),
                        {'num': len(fetched), 'time':
                            time_start_proc_router - time_start_get_router,
                            'time2': time.time() -
                            time_start_proc_router})
            self._process_router_delete()
        except Exception as e:
            LOG.exception(_("Failed synchronizing routers: %s"), str(e))
            self.fullsync = True

    def _process_router_delete(self):
        current_removed_routers = list(self.removed_routers)
        for router_id in current_removed_routers:
            self.removed_routers.remove(router_id)
            if self.servers:
                if not self.servers.rest_delete_router(router_id):
                    self.fail_delete_rest_router_id.add(router_id)

    def _router_ids(self):
        return [self.conf.router_id]


class ZenicAgentWithStateReport(ZenicAgent):

    def __init__(self, host, conf=None):
        super(ZenicAgentWithStateReport, self).__init__(host=host, conf=conf)
        self.state_rpc = agent_rpc.PluginReportStateAPI(topics.PLUGIN)
        self.agent_state = {
            'binary': 'neutron_zenic_agent',
            'host': host,
            'topic': topics.L3_AGENT,
            'configurations': {'agent_mode': self.conf.agent_mode, },
            'start_flag': True,
            'agent_type': 'L3 agent'}
        self.use_call = True
        self.heartbeat = loopingcall.FixedIntervalLoopingCall(
            self._report_state)
        self.heartbeat.start(30)

    def _report_state(self):
        LOG.debug("Report state task started")
        # configurations = self.agent_state['configurations']
        try:
            self.state_rpc.report_state(self.context, self.agent_state,
                                        self.use_call)
            self.agent_state.pop('start_flag', None)
            self.use_call = False
            LOG.debug("Report state task successfully completed")
        except AttributeError:
            # This means the server does not support report_state
            LOG.warn("Neutron server does not support state report."
                     "State report for this agent will be disabled.")
            self.heartbeat.stop()
            return
        except Exception:
            LOG.exception("Failed reporting state!")

    def agent_updated(self, context, payload):
        """Handle the agent_updated notification event."""
        self.fullsync = True
        LOG.info(_("agent_updated by server side %s!"), payload)


def main(manager='networking_zte.l3.zenic_agent.ZenicAgentWithStateReport'):
    register_opts(cfg.CONF)
    common_config.init(sys.argv[1:])
    config.setup_logging()
    server = neutron_service.Service.create(
        binary='neutron-zenic-agent',
        topic=topics.L3_AGENT,
        report_interval=cfg.CONF.AGENT.report_interval,
        manager=manager)
    try:
        service.launch(server).wait()
    except Exception:
        service.launch(cfg.CONF, server).wait()



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2017 ZTE, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_zte.ml2.common.rest.ml2_zenic \
    import config as pl_config
from networking_zte.ml2.common.rest.ml2_zenic \
    import ml2_zenic_restconf as restconf
from oslo_log import log


LOG = log.getLogger(__name__)


class ZenicL2Driver(object):
    """Mechanism Driver for Zenic Networks Controller.

    This driver relays the network create, update, delete
    operations to the Zenic Controller.
    """

    def __init__(self):
        LOG.debug(_('Initializing zte zenic ml2 driver...'))
        # register plugin config opts
        pl_config.register_config()
        # backend doesn't support bulk operations yet
        self.native_bulk_support = False
        # init network ctrl connections
        self.servers = restconf.ZenicServerPool(
            pl_config.cfg.CONF.RESTPROXY.servers,
            pl_config.cfg.CONF.RESTPROXY.server_auth,
            pl_config.cfg.CONF.RESTPROXY.zenic_version,
            pl_config.cfg.CONF.RESTPROXY.enable_qos,
            pl_config.cfg.CONF.RESTPROXY.server_ssl,
            pl_config.cfg.CONF.RESTPROXY.no_ssl_validation,
            pl_config.cfg.CONF.RESTPROXY.ssl_sticky,
            pl_config.cfg.CONF.RESTPROXY.ssl_cert_directory,
            pl_config.cfg.CONF.RESTPROXY.consistency_interval,
            pl_config.cfg.CONF.RESTPROXY.server_timeout,
            pl_config.cfg.CONF.RESTPROXY.cache_connections)
        LOG.debug(_("Initialize zte zenic ml2 driver is done!"))

    def set_enable_security_group(self, en_security_group):
        self.servers.set_enable_security_group(en_security_group)

    def create_network(self, mech_context):
        # create network on the network controller
        self.servers.rest_create_network(mech_context)

    def update_network(self, mech_context):
        # update network on the network controller
        self.servers.rest_update_network(mech_context)

    def delete_network(self, mech_context):
        # delete network on the network controller
        self.servers.rest_delete_network(mech_context)

    def create_subnet(self, mech_context):
        # create subnet on the network controller
        self.servers.rest_create_subnet(mech_context)

    def update_subnet(self, mech_context):
        # update subnet on the network controller
        self.servers.rest_update_subnet(mech_context)

    def delete_subnet(self, mech_context):
        # delete subnet on the network controller
        self.servers.rest_delete_subnet(mech_context)

    def create_port(self, mech_context):
        # create port on the network controller
        self.servers.rest_create_port(mech_context)

    def update_port(self, mech_context):
        # update port on the network controller
        self.servers.rest_update_port(mech_context)

    def delete_port(self, mech_context):
        # delete port on the network controller
        self.servers.rest_delete_port(mech_context)

    def create_security_group(self, mech_context):
        # create security group on the network controller
        self.servers.rest_create_security_group(mech_context)

    def update_security_group(self, mech_context):
        # update security group on the network controller
        self.servers.rest_update_security_group(mech_context)

    def delete_security_group(self, mech_context):
        # delete security group on the network controller
        self.servers.rest_delete_security_group(mech_context)

    def create_security_group_rule(self, mech_context):
        # create securitygroup rule on the network controller
        self.servers.rest_create_security_group_rule(mech_context)

    def update_security_group_rule(self, mech_context):
        # update securitygroup rule on the network controller
        self.servers.rest_update_security_group_rule(mech_context)

    def delete_security_group_rule(self, mech_context):
        # delete securitygroup rule on the network controller
        self.servers.rest_delete_security_group_rule(mech_context)

    def create_qos_policy(self, mech_context):
        self.servers.rest_create_qos_policy(mech_context)

    def update_qos_policy(self, mech_context):
        self.servers.rest_update_qos_policy(mech_context)

    def delete_qos_policy(self, mech_context):
        self.servers.rest_delete_qos_policy(mech_context)

    def create_qos_bandwidth_limit_rule(self, mech_context):
        self.servers.rest_create_qos_bandwidth_limit_rule(mech_context)

    def update_qos_bandwidth_limit_rule(self, mech_context):
        self.servers.rest_update_qos_bandwidth_limit_rule(mech_context)

    def delete_qos_bandwidth_limit_rule(self, mech_context):
        self.servers.rest_delete_qos_bandwidth_limit_rule(mech_context)



# Copyright 2015 ZTE, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from abc import ABCMeta
from abc import abstractproperty
import six


@six.add_metaclass(ABCMeta)
class SecurityGroupContext(object):
    """Context passed to MechanismDrivers for changes to
    security_group resources.

    A SecurityGroupContext instance wraps a security_group resource.
    It provides helper methods for accessing other relevant information.
    Results from expensive operations are cached so that other
    MechanismDrivers can freely access the same information.
    """

    @abstractproperty
    def current(self):
        """Return the current state of the security_group.

        Return the current state of the security_group, as defined by
        NeutronPluginBaseV2.create_security_group and all extensions in the
        ml2 plugin.
        """
        pass

    @abstractproperty
    def original(self):
        """Return the original state of the security_group.

        Return the original state of the security_group, prior to a call to
        update_security_group. Method is only valid within calls to
        update_security_group_precommit and update_security_group_postcommit.
        """
        pass


@six.add_metaclass(ABCMeta)
class SecurityGroupRuleContext(object):
    """Context passed to MechanismDrivers for changes
    to security_group_rule resources.

    A SecurityGroupRuleContext instance wraps a
    security_group_rule resource. It provides
    helper methods for accessing other relevant information. Results
    from expensive operations are cached so that other
    MechanismDrivers can freely access the same information.
    """

    @abstractproperty
    def current(self):
        """Return the current state of the security_group_rule.

        Return the current state of the security_group_rule, as defined by
        NeutronPluginBaseV2.create_security_group_rule and all extensions
        in the  ml2 plugin.
        """
        pass

    @abstractproperty
    def original(self):
        """Return the original state of the subnet.

        Return the original state of the security_group_rule, prior to a call
        to update_security_group_rule. Method is only valid within calls to
        update_security_group_rule_precommit and
        update_security_group_rule_postcommit.
        """
        pass


@six.add_metaclass(ABCMeta)
class QosPolicyContext(object):
    @abstractproperty
    def current(self):
        pass

    @abstractproperty
    def original(self):
        pass


@six.add_metaclass(ABCMeta)
class QosBandWidthRuleContext(object):

    @abstractproperty
    def current(self):
        pass

    @abstractproperty
    def original(self):
        pass



# Copyright 2011 ZTE, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_zte.common import servermanager
from networking_zte.ml2.common.rest.ml2_zenic import config as pl_config
from networking_zte.ml2 import driver_context
from networking_zte.ml2 import sync_filter
try:
    from neutron import context as neutron_context
except Exception:
    from neutron_lib import context as neutron_context
from neutron.db import db_base_plugin_v2
from neutron.db import securitygroups_rpc_base as sg_db_rpc
try:
    from neutron_lib.api.definitions import portbindings
except Exception:
    from neutron.extensions import portbindings
from neutron.extensions import securitygroup as ext_sg
from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron.plugins.ml2 import driver_api as api
from neutron.plugins.ml2 import managers as manager
from neutron.plugins.ml2 import plugin
from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils
from oslo_utils import importutils
import six

LOG = log.getLogger(__name__)

ML2_NETWORK = 'network'
ML2_NETWORKS = 'networks'
ML2_SUBNET = 'subnet'
ML2_SUBNETS = 'subnets'
ML2_PORT = 'port'
ML2_PORTS = 'ports'
ML2_SG = 'security_group'
ML2_SGS = 'security_groups'
ML2_SG_RULE = 'security_group_rule'
ML2_SG_RULES = 'security_group_rules'


proxy_opts = [
    cfg.StrOpt(
        'driver_list',
        default='sdn:networking_zte.ml2.'
                'common.rest.ml2_zenic.mech_zenic_l2.ZenicL2Driver',
        help="Define the drivers of proxy agent"),
]
security_group_opts = [
    cfg.BoolOpt(
        'enable_security_group',
        default=True,
        help=_(
            'Controls whether the neutron security group API is enabled '
            'in the server. It should be false when using no security '
            'groups or using the nova security group API.')),
]
CONF = cfg.CONF
CONF.register_opts(proxy_opts, 'ml2')
CONF.register_opts(security_group_opts, 'securitygroup')


def method_patch(cls, bases, dct):
    base = bases[0]
    for name, value in dct.iteritems():
        if name not in("__module__", "__metaclass__", "__doc__"):
            setattr(base, name, value)
    return base


@six.add_metaclass(method_patch)
class MechanismDriverExt(api.MechanismDriver):
    """Define stable abstract interface extension for ML2 mechanism
    drivers.
    """

    def bind_port(self, context):
        binding_levels = context.binding_levels
        if hasattr(context, "vif_type"):
            vif_type = context.vif_type
        else:
            LOG.info(_("zte_ml2 bind_port vif_type is None"))
            return
        LOG.info(_("zte_ml2 bind_port binding_level =%s"), binding_levels)

        LOG.info(_("zte_ml2 bind_port binding_level =%s"), binding_levels)
        if (binding_levels is not None and
                binding_levels['driver'] == 'ml2_ztedriver') or \
                (vif_type != portbindings.VIF_TYPE_UNBOUND and
                    vif_type != portbindings.VIF_TYPE_BINDING_FAILED):
            return
        pl_config.register_config()
        if pl_config.cfg.CONF.RESTPROXY.enable_hierarchical_port == "True":
            LOG.info(_("zte_ml2 bind_port segments_to_bind= %s"),
                     context.segments_to_bind)
            next_segments_to_bind = list()
            segment_to_bind = context.segments_to_bind[-1]
            physical_network = \
                segment_to_bind.get(api.PHYSICAL_NETWORK, None)
            if physical_network is None:
                LOG.info("physical_network is not configed!")
            segment = {api.NETWORK_TYPE: 'vlan',
                       api.PHYSICAL_NETWORK: physical_network,
                       }
            LOG.info(_("zte_ml2 bind_port segment= %s"), segment)
            next_segment_to_bind = \
                context.allocate_dynamic_segment(segment)
            LOG.info(_("zte_ml2 bind_port next_segment_to_bind= %s"),
                     next_segment_to_bind)
            next_segments_to_bind.append(next_segment_to_bind)
            next_segments_to_bind.append(segment_to_bind)

            context.continue_binding(segment_to_bind[api.ID],
                                     next_segments_to_bind)

    def create_security_group_precommit(self, context):
        """Create resources of a security group.

        :param context: SecurityGroupContext instance describing the current
        state of the security group, prior to the call to create it.

        Called inside transaction context on session. Runtime errors
        are not expected, but raising an exception will result in
        rollback of the transaction.
        """
        pass

    def create_security_group_postcommit(self, context):
        """Create a security group.

        :param context: SecurityGroupContext instance describing the current
        state of the security group, prior to the call to create it.

        Called after the transaction completes. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance.  Runtime errors are not
        expected, and will not prevent the resource from being
        created.
        """
        pass

    def update_security_group_precommit(self, context):
        """Create resources of a security group.

        :param context: SecurityGroupContext instance describing the current
        state of the security group, prior to the call to create it.

        Called inside transaction context on session. Runtime errors
        are not expected, but raising an exception will result in
        rollback of the transaction.
        """
        LOG.debug("MechanismDriverExt precommit")
        pass

    def update_security_group_postcommit(self, context):
        """Create resources of a security group.

        :param context: SecurityGroupContext instance describing the current
        state of the security group, prior to the call to create it.

        Called inside transaction context on session. Runtime errors
        are not expected, but raising an exception will result in
        rollback of the transaction.
        """
        LOG.debug("MechanismDriverExt postcommit")
        pass

    def delete_security_group_precommit(self, context):
        """Delete resources of a security group.

        :param context: SecurityGroupContext instance describing the current
        state of the security group, prior to the call to delete it.

        Called inside transaction context on session. Runtime errors
        are not expected, but raising an exception will result in
        rollback of the transaction.
        """
        pass

    def delete_security_group_postcommit(self, context):
        """Delete a security group.

        :param context: SecurityGroupContext instance describing the current
        state of the security group, prior to the call to delete it.

        Called after the transaction completes. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance.  Runtime errors are not
        expected, and will not prevent the resource from being
        deleted.
        """
        pass

    def create_security_group_rule_precommit(self, context):
        """Create resources of a security group rule.

        :param context: SecurityGroupContext instance describing the current
        state of the security group, prior to the call to create it.

        Called inside transaction context on session. Runtime errors
        are not expected, but raising an exception will result in
        rollback of the transaction.
        """
        pass

    def create_security_group_rule_postcommit(self, context):
        """Create a security group rule.

        :param context: SecurityGroupContext instance describing the current
        state of the security group, prior to the call to create it.

        Called after the transaction completes. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance.  Runtime errors are not
        expected, and will not prevent the resource from being
        created.
        """
        pass

    def delete_security_group_rule_precommit(self, context):
        """Delete resources of a security group rule.

        :param context: SecurityGroupContext instance describing the current
        state of the security group, prior to the call to delete it.

        Called inside transaction context on session. Runtime errors
        are not expected, but raising an exception will result in
        rollback of the transaction.
        """
        pass

    def delete_security_group_rule_postcommit(self, context):
        """Delete a security group rule.

        :param context: SecurityGroupContext instance describing the current
        state of the security group, prior to the call to delete it.

        Called after the transaction completes. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance.  Runtime errors are not
        expected, and will not prevent the resource from being
        deleted.
        """
        pass

    def create_qos_policy_precommit(self, context):
        pass

    def update_qos_policy_precommit(self, context):
        pass

    def delete_qos_policy_precommit(self, context):
        pass

    def create_qos_bandwidth_limit_rule_precommit(self, context):
        pass

    def update_qos_bandwidth_limit_rule_precommit(self, context):
        pass

    def delete_qos_bandwidth_limit_rule_precommit(self, context):
        pass


@six.add_metaclass(method_patch)
class MechanismManagerExt(manager.MechanismManager):
    """Manage networking mechanisms using drivers."""

    def create_security_group_precommit(self, context):
        """Notify all mechanism drivers during security group creation.

        :raises: neutron.plugins.ml2.common.MechanismDriverError
        if any mechanism driver create_security_group_precommit call fails.

        Called within the database transaction. If a mechanism driver
        raises an exception, then a MechanismDriverError is propogated
        to the caller, triggering a rollback. There is no guarantee
        that all mechanism drivers are called in this case.
        """
        self._call_on_drivers("create_security_group_precommit", context)

    def create_security_group_postcommit(self, context):
        """Notify all mechanism drivers after security group creation.

        :raises: neutron.plugins.ml2.common.MechanismDriverError
        if any mechanism driver create_security_group_postcommit call fails.

        Called after the database transaction. If a mechanism driver
        raises an exception, then a MechanismDriverError is propagated
        to the caller, where the subnet will be deleted, triggering
        any required cleanup. There is no guarantee that all mechanism
        drivers are called in this case.
        """
        self._call_on_drivers("create_security_group_postcommit", context)

    def delete_security_group_precommit(self, context):
        """Notify all mechanism drivers during security group deletion.

        :raises: neutron.plugins.ml2.common.MechanismDriverError
        if any mechanism driver create_security_group_precommit call fails.

        Called within the database transaction. If a mechanism driver
        raises an exception, then a MechanismDriverError is propogated
        to the caller, triggering a rollback. There is no guarantee
        that all mechanism drivers are called in this case.
        """
        self._call_on_drivers("delete_security_group_precommit", context)

    def delete_security_group_postcommit(self, context):
        """Notify all mechanism drivers after security group deletion.

        :raises: neutron.plugins.ml2.common.MechanismDriverError
        if any mechanism driver create_security_group_postcommit call fails.

        Called after the database transaction. If a mechanism driver
        raises an exception, then a MechanismDriverError is propagated
        to the caller, where the subnet will be deleted, triggering
        any required cleanup. There is no guarantee that all mechanism
        drivers are called in this case.
        """
        self._call_on_drivers("delete_security_group_postcommit", context)

    def update_security_group_precommit(self, context):
        """Notify all mechanism drivers during security group creation.

        :raises: neutron.plugins.ml2.common.MechanismDriverError
        if any mechanism driver update_security_group_precommit call fails.

        Called within the database transaction. If a mechanism driver
        raises an exception, then a MechanismDriverError is propogated
        to the caller, triggering a rollback. There is no guarantee
        that all mechanism drivers are called in this case.
        """
        LOG.debug("MechanismManagerExt precommit")
        self._call_on_drivers("update_security_group_precommit", context)

    def update_security_group_postcommit(self, context):
        """Notify all mechanism drivers during security group creation.

        :raises: neutron.plugins.ml2.common.MechanismDriverError
        if any mechanism driver update_security_group_postcommit call fails.

        Called within the database transaction. If a mechanism driver
        raises an exception, then a MechanismDriverError is propogated
        to the caller, triggering a rollback. There is no guarantee
        that all mechanism drivers are called in this case.
        """
        LOG.debug("MechanismManagerExt postcommit")
        self._call_on_drivers("update_security_group_postcommit", context)

    def create_security_group_rule_precommit(self, context):
        """Notify all mechanism drivers during security group rule creation.

        :raises: neutron.plugins.ml2.common.MechanismDriverError
        if any mechanism driver create_security_group_rule_precommit
        call fails.

        Called within the database transaction. If a mechanism driver
        raises an exception, then a MechanismDriverError is propogated
        to the caller, triggering a rollback. There is no guarantee
        that all mechanism drivers are called in this case.
        """
        self._call_on_drivers("create_security_group_rule_precommit", context)

    def create_security_group_rule_postcommit(self, context):
        """Notify all mechanism drivers after security group rule creation.

        :raises: neutron.plugins.ml2.common.MechanismDriverError
        if any mechanism driver create_security_group_rule_postcommit
        call fails.

        Called after the database transaction. If a mechanism driver
        raises an exception, then a MechanismDriverError is propagated
        to the caller, where the subnet will be deleted, triggering
        any required cleanup. There is no guarantee that all mechanism
        drivers are called in this case.
        """
        self._call_on_drivers("create_security_group_rule_postcommit", context)

    def delete_security_group_rule_precommit(self, context):
        """Notify all mechanism drivers during security group rule deletion.

        :raises: neutron.plugins.ml2.common.MechanismDriverError
        if any mechanism driver create_security_group_rule_precommit
        call fails.

        Called within the database transaction. If a mechanism driver
        raises an exception, then a MechanismDriverError is propogated
        to the caller, triggering a rollback. There is no guarantee
        that all mechanism drivers are called in this case.
        """
        self._call_on_drivers("delete_security_group_rule_precommit", context)

    def delete_security_group_rule_postcommit(self, context):
        """Notify all mechanism drivers after security group rule deletion.

        :raises: neutron.plugins.ml2.common.MechanismDriverError
        if any mechanism driver create_security_group_rule_postcommit
        call fails.

        Called after the database transaction. If a mechanism driver
        raises an exception, then a MechanismDriverError is propagated
        to the caller, where the subnet will be deleted, triggering
        any required cleanup. There is no guarantee that all mechanism
        drivers are called in this case.
        """
        self._call_on_drivers("delete_security_group_rule_postcommit", context)

    def create_qos_policy_precommit(self, context):
        self._call_on_drivers("create_qos_policy_precommit", context)

    def update_qos_policy_precommit(self, context):
        self._call_on_drivers("update_qos_policy_precommit", context)

    def delete_qos_policy_precommit(self, context):
        self._call_on_drivers("delete_qos_policy_precommit", context)

    def create_qos_bandwidth_limit_rule_precommit(self, context):
        self._call_on_drivers(
            "create_qos_bandwidth_limit_rule_precommit", context)

    def update_qos_bandwidth_limit_rule_precommit(self, context):
        self._call_on_drivers(
            "update_qos_bandwidth_limit_rule_precommit", context)

    def delete_qos_bandwidth_limit_rule_precommit(self, context):
        self._call_on_drivers(
            "delete_qos_bandwidth_limit_rule_precommit", context)


@six.add_metaclass(method_patch)
class Ml2PluginExt(plugin.Ml2Plugin):
    """Extend the Neutron L2 abstractions using modules.

    Implement security_group,security_group_rule methods to notify
    SDN controller.
    """

    def create_qos(self, context, qos):
        LOG.info(_("Ml2PluginExt create_qos qos = %s"), qos)
        result = super(Ml2PluginExt, self).create_qos(context, qos)

        rate_limit = result['policies']['ratelimit']
        bandwidth_limit_rule = {
            "tenant_id": result['tenant_id'],
            "id": result['id'],
            "max_kbps": rate_limit,
        }
        mech_bandwidth_context = driver_context.QosBandWidthRuleContext(
            self, context, bandwidth_limit_rule)
        mech_policy_context = driver_context.QosPolicyContext(
            self, context, result)
        try:
            self.mechanism_manager.create_qos_policy_precommit(
                mech_policy_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error(_("mechanism_manager."
                            "create_qos_precommit failed, "
                            "create qos '%s'"), result)
                super(Ml2PluginExt, self).delete_qos(context, result['id'])

        try:
            self.mechanism_manager.create_qos_bandwidth_limit_rule_precommit(
                mech_bandwidth_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error(_("mechanism_manager."
                            "create_qos_precommit failed, "
                            "create qos '%s'"), result)
                self.mechanism_manager.delete_qos_policy_precommit(
                    mech_policy_context)
                super(Ml2PluginExt, self).delete_qos(context, result['id'])

        return result

    def update_qos(self, context, id, qos):
        qos_info = qos['qos']
        LOG.info(_("Ml2PluginExt update_qos qos = %s"), qos_info)
        if 'policies' in qos_info:
            rate_limit = qos_info['policies']['ratelimit']
            bandwidth_limit_rule = {
                "id": id,
                "max_kbps": rate_limit,
            }
            mech_bandwidth_context = driver_context.QosBandWidthRuleContext(
                self, context, bandwidth_limit_rule)
            try:
                self.mechanism_manager.update_qos_bandwidth_limit_rule_precommit(        # noqa
                    mech_bandwidth_context)
            except ml2_exc.MechanismDriverError:
                with excutils.save_and_reraise_exception():
                    LOG.error(_("mechanism_manager."
                                "update_qos_bandwidth_limit_rule_precommit "
                                "failed, udpate_qos_bandwidth rule '%s'"),
                              bandwidth_limit_rule)

        if 'description' in qos_info:
            policy = {
                "id": id,
                "description": qos_info['description'],
            }
            mech_policy_context = driver_context.QosPolicyContext(
                self, context, policy)
            try:
                self.mechanism_manager.update_qos_policy_precommit(
                    mech_policy_context)
            except ml2_exc.MechanismDriverError:
                with excutils.save_and_reraise_exception():
                    LOG.error(_("mechanism_manager."
                                "update_qos_policy_precommit failed, "
                                "update_qos_policy '%s'"), policy)

        result = super(Ml2PluginExt, self).update_qos(context,
                                                      id,
                                                      qos)
        return result

    def delete_qos(self, context, id):
        bandwidth_limit_rule = {
            "id": id,
        }
        mech_bandwidth_context = driver_context.QosBandWidthRuleContext(
            self, context, bandwidth_limit_rule)

        policy = {
            "id": id,
        }
        mech_policy_context = driver_context.QosPolicyContext(
            self, context, policy)

        try:
            self.mechanism_manager.delete_qos_policy_precommit(
                mech_policy_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error(_("mechanism_manager."
                            "delete_qos_policy_precommit failed, "
                            "deleting qos '%s'"), policy)
        try:
            self.mechanism_manager.delete_qos_bandwidth_limit_rule_precommit(
                mech_bandwidth_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error(_("mechanism_manager."
                            "delete_qos_bandwidth_limit_rule_precommit "
                            "failed, deleting qos '%s'"), policy)
        super(Ml2PluginExt, self).delete_qos(context, id)

    def extend_port_extra_dhcp_opts_dict(self, context, port):
        self._extend_port_extra_dhcp_opts_dict(context, port)

    def create_security_group(
            self, context, security_group, default_sg=False):
        """Create security group.

        Create a new security group, including the default security group.
        """
        LOG.debug(_("Ml2Plugin.create_security_group called: "
                    "security_group=%(security_group)s "
                    "default_sg=%(default_sg)s "),
                  {'security_group': security_group,
                   'default_sg': default_sg})

        with context.session.begin(subtransactions=True):
            sg = security_group.get('security_group')
            tenant_id = sg['tenant_id']
            if not default_sg:
                self._ensure_default_security_group(context, tenant_id)

            # Create the Neutron sg first
            sg = super(plugin.Ml2Plugin, self).create_security_group(
                context, security_group, default_sg)

            LOG.debug(_("sg: %s"), sg)

            mech_context = driver_context.SecurityGroupContext(
                self, context, sg)
            self.mechanism_manager.create_security_group_precommit(
                mech_context)

        try:
            self.mechanism_manager.create_security_group_postcommit(
                mech_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error(_("mechanism_manager."
                            "create_security_group_postcommit failed, "
                            "deleting security group '%s'"), sg)
                self.delete_security_group(context, sg['id'])

        return sg

    def update_security_group(
            self, context, id, security_group):
        """Update security group.

        update security group, not including the default security group.
        """

        LOG.debug(_("update_sg,context= %s"), (', '.join(
            ['%s:%s' % item for item in context.__dict__.items()])))
        LOG.debug(_("Ml2Plugin.update_security_group called: "
                    "security_group=%(security_group)s "),
                  {'security_group': security_group})

        sg = super(plugin.Ml2Plugin, self).update_security_group(
            context, id, security_group)

        LOG.debug(_("sg: %s"), sg)

        mech_context = driver_context.SecurityGroupContext(
            self, context, sg)
        self.mechanism_manager.update_security_group_precommit(
            mech_context)

        try:
            self.mechanism_manager.update_security_group_postcommit(
                mech_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error(_("mechanism_manager."
                            "update_security_group_postcommit failed, "
                            "deleting security group '%s'"), sg)

        return sg

    def delete_security_group(self, context, id):
        """Delete chains for Neutron security group."""
        LOG.debug(_("Ml2Plugin.delete_security_group called: id=%s"), id)

        with context.session.begin(subtransactions=True):
            sg = super(plugin.Ml2Plugin, self).get_security_group(context, id)

            if not sg:
                raise ext_sg.SecurityGroupNotFound(id=id)

            if sg["name"] == 'default' and not context.is_admin:
                raise ext_sg.SecurityGroupCannotRemoveDefault()

            sg_id = sg['id']
            filters = {'security_group_id': [sg_id]}
            if super(
                plugin.Ml2Plugin, self)._get_port_security_group_bindings(
                    context, filters):
                raise ext_sg.SecurityGroupInUse(id=sg_id)

            super(plugin.Ml2Plugin, self).delete_security_group(context, id)
            mech_context = driver_context.SecurityGroupContext(
                self, context, sg)

            self.mechanism_manager.delete_security_group_precommit(
                mech_context)

        try:
            self.mechanism_manager.delete_security_group_postcommit(
                mech_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error(_("mechanism_manager."
                            "delete_security_group_postcommit "
                            "failed, deleting security group '%s'"), sg)

    def create_security_group_rule(self, context, security_group_rule):
        """Create a security group rule

        Create a security group rule in the Neutron DB and corresponding
        SDN resources in its data store.
        """
        LOG.debug(_("Ml2Plugin.create_security_group_rule called: "
                    "security_group_rule=%(security_group_rule)r"),
                  {'security_group_rule': security_group_rule})

        with context.session.begin(subtransactions=True):
            rule = super(plugin.Ml2Plugin, self).create_security_group_rule(
                context, security_group_rule)

            LOG.debug(_("Ml2Plugin.create_security_group_rule exiting: "
                        "rule=%r"), rule)

            mech_context = driver_context.SecurityGroupRuleContext(
                self, context, rule)
            self.mechanism_manager.create_security_group_rule_precommit(
                mech_context)

        try:
            self.mechanism_manager.create_security_group_rule_postcommit(
                mech_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error(_("mechanism_manager."
                            "create_security_group_rule_postcommit "
                            "failed, security group rule'%s'"), rule)
                self.delete_security_group_rule(context, rule["id"])
        return rule

    def create_security_group_rule_bulk(self, context, security_group_rules):
        created_rules = list()
        for security_group_rule in security_group_rules[
            'security_group_rules']:
            created_rules.append(
                self.create_security_group_rule(
                    context, security_group_rule))
        return created_rules

    def delete_security_group_rule(self, context, sg_rule_id):
        """Delete a security group rule

        Delete a security group rule from the Neutron DB and corresponding
        SDN resources from its data store.
        """
        LOG.debug(_("Ml2Plugin.delete_security_group_rule called: "
                    "sg_rule_id=%s"), sg_rule_id)
        with context.session.begin(subtransactions=True):
            rule = super(plugin.Ml2Plugin, self).get_security_group_rule(
                context, sg_rule_id)

            if not rule:
                raise ext_sg.SecurityGroupRuleNotFound(id=sg_rule_id)

            super(plugin.Ml2Plugin, self).delete_security_group_rule(
                context, sg_rule_id)

            mech_context = driver_context.SecurityGroupRuleContext(
                self, context, rule)
            self.mechanism_manager.delete_security_group_rule_precommit(
                mech_context)

        try:
            self.mechanism_manager.delete_security_group_rule_postcommit(
                mech_context)
        except ml2_exc.MechanismDriverError:
            with excutils.save_and_reraise_exception():
                LOG.error(_("mechanism_manager."
                            "delete_security_group_rule_postcommit "
                            "failed, security group rule'%s'"), rule)


class Ml2MechanismDriver(api.MechanismDriver,
                         db_base_plugin_v2.NeutronDbPluginV2,
                         sg_db_rpc.SecurityGroupServerRpcMixin):

    """Mechanism Driver for Ml2Plugin.
    """
    FILTER_MAP = {
        ML2_NETWORK: sync_filter.NetworkFilter,
        ML2_SUBNET: sync_filter.SubnetFilter,
        ML2_PORT: sync_filter.PortFilter,
        ML2_SG: sync_filter.SecurityGroupFilter,
        ML2_SG_RULE: sync_filter.SecurityGroupRuleFilter,
    }

    def initialize(self):
        self.driver_provider = {}
        self._load_driver_provider()
        self.set_enable_security_group()
        pl_config.register_config()
        self.enable_pre_commit = pl_config.cfg.CONF.RESTPROXY.enable_pre_commit
        LOG.info(_(
            "ZTE Ml2MechanismDriver enable_pre_commit = %s"),
            self.enable_pre_commit)

    def sync_full(self, context=None):
        dbcontext = neutron_context.get_admin_context()
        for collection_name in [ML2_NETWORK,
                                ML2_SUBNET,
                                ML2_PORT,
                                ML2_SG,
                                ML2_SG_RULE]:
            filter_cls = self.FILTER_MAP[collection_name]
            obj_getter = getattr(self, 'get_%ss' % collection_name)
            resources = obj_getter(dbcontext)
            for resource in resources:
                LOG.debug(
                    "sync_full sync %s = %s" %
                    (collection_name, resource))
                operation = 'create' + '_' + collection_name
                if context is not None:
                    mech_context = \
                        filter_cls.filter_create_attributes_with_plugin(
                            resource, context._plugin, dbcontext)
                else:
                    if 'network_id' in resource:
                        network = self.get_network(
                            dbcontext, resource['network_id'])
                        LOG.info(_("sync_full,network= %s"), network)
                    else:
                        network = None
                    mech_context = \
                        filter_cls.filter_create_attributes_without_plugin(
                            resource, dbcontext, network)

                try:
                    self._call_on_drivers_to_rest(operation, mech_context)
                    LOG.debug(_("sync_full success!,resource= %s"),
                              (', '.join(['%s:%s' %
                         item for item in mech_context.__dict__.items()])))
                except BaseException as e:
                    LOG.error(_("sync_full error!,resource= %s"), (', '.join(
                        ['%s:%s' %
                         item for item in mech_context.__dict__.items()])))
                    LOG.error(_("sync_full, except:%s"), str(e))

    def _load_driver_provider(self):
        """Loads service drivers.
        """
        driver_providers = [driver_set.split(':')
                            for driver_set in CONF.ml2.driver_list.split(',')]

        while [''] in driver_providers:
            driver_providers.remove([''])

        if not driver_providers:
            LOG.info(_("Can't find service drivers"))
            return

        LOG.debug(_("Loading service drivers: %s"), driver_providers)
        for driver_name, driver_class in driver_providers:
            if driver_name is None or driver_class is None:
                continue

            LOG.debug(_("Loading service driver: %(dn)s, class path: %(cla)s"),
                      {"dn": driver_name, "cla": driver_class})
            driver_cls = importutils.import_class(driver_class)
            driver_inst = driver_cls()

            self.driver_provider[driver_name] = driver_inst

            LOG.info(_("Successfully loaded %(name)s driver. "
                       "Description: %(desc)s"),
                     {"name": driver_name, "desc": driver_class})

    def _call_on_drivers(self, method, context, continue_on_failure=False):
        """Call method of service drivers.
        """
        self._call_on_drivers_to_rest(method, context, continue_on_failure)
#         try:
#             self._call_on_drivers_to_rest(method,context,continue_on_failure)
#         except:
#             self.sync_full(context)
#             self._call_on_drivers_to_rest(method,context,
#             continue_on_failure)

    def _call_on_drivers_to_rest(
            self,
            method,
            context,
            continue_on_failure=False):
        """Call method of service drivers.
        """
        if not self.driver_provider:
            return

        error = False
        for name, inst in self.driver_provider.items():
            try:
                if hasattr(inst, method):
                    getattr(inst, method)(context)
            except servermanager.RemoteRestError:
                LOG.error(
                    _("ZTE Ml2MechanismDriver '%(name)s' failed in "
                      "%(method)s"), {'name': name, 'method': method}
                )
                error = True
                if not continue_on_failure:
                    break

        if error:
            raise ml2_exc.MechanismDriverError(
                method=method
            )

    def set_enable_security_group(self):
        en_security_group = CONF.securitygroup.enable_security_group
        self._call_on_drivers_to_rest(
            'set_enable_security_group', en_security_group)
    # Postcommit hooks are used to trigger synchronization.

    def create_network_postcommit(self, context):
        if CONF.securitygroup.enable_security_group:
            tenant_id = context.current.get("tenant_id", None)
            if tenant_id:
                default_sg_id = self._ensure_default_security_group(
                    context._plugin_context, tenant_id)
                default_sg = self.get_security_group(
                    context._plugin_context, default_sg_id)

                LOG.debug(_("ZTE Ml2MechanismDriver."
                            "create_network_postcommit called: "
                            "security_group=%(security_group)s "
                            "default_sg_id=%(default_sg)s "),
                          {'security_group': default_sg,
                           'default_sg': default_sg_id})
                mech_context = driver_context.SecurityGroupContext(
                    self, context._plugin_context, default_sg)

                try:
                    self.create_security_group_postcommit(mech_context)
                except ml2_exc.MechanismDriverError:
                    with excutils.save_and_reraise_exception():
                        LOG.error(_("ZTE Ml2MechanismDriver."
                                    "create_security_group_postcommit "
                                    "failed '%s'"), default_sg)
        self._call_on_drivers('create_network', context)

    def delete_network_precommit(self, context):
        self._call_on_drivers_to_rest('delete_network', context)

    def update_network_precommit(self, context):
        if self.enable_pre_commit == "True":
            LOG.debug("ZTE Ml2MechanismDriver enter update_subnet_precommit")
            self._call_on_drivers('update_network', context)

    def update_network_postcommit(self, context):
        if self.enable_pre_commit == "False":
            LOG.debug("ZTE Ml2MechanismDriver enter update_network_postcommit")
            self._call_on_drivers('update_network', context)

    def create_subnet_postcommit(self, context):
        self._call_on_drivers('create_subnet', context)

    def delete_subnet_precommit(self, context):
        self._call_on_drivers_to_rest('delete_subnet', context)

    def update_subnet_postcommit(self, context):
        if self.enable_pre_commit == "False":
            LOG.debug("ZTE Ml2MechanismDriver enter update_subnet_postcommit")
            self._call_on_drivers('update_subnet', context)

    def update_subnet_precommit(self, context):
        if self.enable_pre_commit == "True":
            LOG.debug("ZTE Ml2MechanismDriverenter update_subnet_precommit")
            self._call_on_drivers('update_subnet', context)

    def create_port_postcommit(self, context):
        if CONF.securitygroup.enable_security_group:
            tenant_id = context.current.get("tenant_id", None)
            if tenant_id:
                default_sg_id = self._ensure_default_security_group(
                    context._plugin_context, tenant_id)
                default_sg = self.get_security_group(
                    context._plugin_context, default_sg_id)

                LOG.debug(_("ZTE Ml2MechanismDriver."
                            "create_network_postcommit called: "
                            "security_group=%(security_group)s "
                            "default_sg_id=%(default_sg)s "),
                          {'security_group': default_sg,
                           'default_sg': default_sg_id})
                mech_context = driver_context.SecurityGroupContext(
                    self, context._plugin_context, default_sg)

                try:
                    self.create_security_group_postcommit(mech_context)
                except ml2_exc.MechanismDriverError:
                    with excutils.save_and_reraise_exception():
                        LOG.error(_("ZTE Ml2MechanismDriver."
                                    "create_security_group_postcommit "
                                    "failed '%s'"), default_sg)
        self._call_on_drivers('create_port', context)

    def delete_port_precommit(self, context):
        self._call_on_drivers_to_rest('delete_port', context)

    def update_port_postcommit(self, context):
        if self.enable_pre_commit == "False":
            LOG.debug("ZTE Ml2MechanismDriver enter update_port_postcommit")
            self._call_on_drivers('update_port', context)

    def update_port_precommit(self, context):
        if self.enable_pre_commit == "True":
            LOG.debug("ZTE Ml2MechanismDriver enter update_port_precommit")
            self._call_on_drivers('update_port', context)

    def create_security_group_postcommit(self, context):
        self._call_on_drivers('create_security_group', context)

    def update_security_group_precommit(self, context):
        if self.enable_pre_commit == "True":
            LOG.debug("ZTE Ml2MechanismDriver enter update_security_group_"
                      "precommit")
            self._call_on_drivers('update_security_group', context)

    def update_security_group_postcommit(self, context):
        if self.enable_pre_commit == "False":
            LOG.debug("ZTE Ml2MechanismDriver enter update_security_group_"
                      "postcommit")
            self._call_on_drivers('update_security_group', context)

    def delete_security_group_precommit(self, context):
        self._call_on_drivers_to_rest('delete_security_group', context)

    def create_security_group_rule_postcommit(self, context):
        self._call_on_drivers('create_security_group_rule', context)

    def delete_security_group_rule_precommit(self, context):
        self._call_on_drivers_to_rest('delete_security_group_rule', context)

    def create_qos_policy_precommit(self, context):
        self._call_on_drivers("create_qos_policy", context)

    def update_qos_policy_precommit(self, context):
        self._call_on_drivers("update_qos_policy", context)

    def delete_qos_policy_precommit(self, context):
        self._call_on_drivers_to_rest("delete_qos_policy", context)

    def create_qos_bandwidth_limit_rule_precommit(self, context):
        self._call_on_drivers("create_qos_bandwidth_limit_rule", context)

    def update_qos_bandwidth_limit_rule_precommit(self, context):
        self._call_on_drivers("update_qos_bandwidth_limit_rule", context)

    def delete_qos_bandwidth_limit_rule_precommit(self, context):
        self._call_on_drivers_to_rest(
            "delete_qos_bandwidth_limit_rule", context)

    def check_vlan_transparency(self, context):
        return True

    def notify_sg_on_port_change(self, **kwargs):
        pass



# Copyright 2017 ZTE, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_zte.common import servermanager
from neutron.db import common_db_mixin as base_db
from oslo_log import log

LOG = log.getLogger(__name__)

# The following are used to invoke the API on the external controller

QOS_POLICY = 'qos-policy'
QOS_BANDWIDTH_RULE = 'qos-bandwidth-limit-rule'
QOS_DSCP_RULE = 'qos-dscp-marking-rule'
BASE_URI = '/restconf/operations/zenic-vdcapp-model:'
SUCCESS_CODES = range(200, 207)
FAILURE_CODES = [0, 9, 301, 302, 303, 400, 401, 403, 404, 409, 500, 501, 502,
                 503, 504, 505]


class ZenicRestServerPool(servermanager.ServerPool,
                          base_db.CommonDbMixin):
    """ZenicRestServerPool for qos plugin.
    This server pool has qos_policy and qos_bandwidth operations
    of create, update and delete, to the Zenic Controller.
    """

    def __init__(self, servers, auth, ssl, no_ssl_validation, ssl_sticky,
                 ssl_cert_directory, consistency_interval, timeout=False,
                 cache_connections=True, base_uri=BASE_URI,
                 success_codes=SUCCESS_CODES,
                 failure_codes=FAILURE_CODES, name='ZenicRestProxy'):
        super(ZenicRestServerPool, self).__init__(
            servers, auth, ssl, no_ssl_validation, ssl_sticky,
            ssl_cert_directory, consistency_interval, timeout,
            cache_connections, base_uri, success_codes, failure_codes, name)

    @staticmethod
    def construct_qos_policy_info(context, id, qos_policy, action):
        """
        #    {u'policy': {
        #               'shared': False,
        #               'tenant_id': u'c414cf57e9fc49efb8e3e0c8d4e41b33',
        #                u'name': u'qos-policy',
        #               'description': '',
        #               'rules':[],
        #               'id' :'5720800f-dece-4c9a-93d4-5dc3ef94ccb7'}}
        """
        if action == 'UPDATE':
            qos_policy = qos_policy['policy']
        qos_policy_info = {
            "input": {
                "id": id if (id != 0) else qos_policy['id'],
            }
        }
        input = qos_policy_info['input']
        if 'type' in qos_policy:
            input['type'] = qos_policy['type']
        if 'description' in qos_policy:
            input['description'] = qos_policy['description']
        if 'tenant_id' in qos_policy:
            input['tenant_id'] = qos_policy['tenant_id']
        if 'shared' in qos_policy:
            input['shared'] = qos_policy['shared']
        if 'name' in qos_policy:
            input['name'] = qos_policy['name']

        return qos_policy_info

    @staticmethod
    def construct_bandwidth_limit_rule_info(context, rule_id, policy_id,
                                            bandwidth_limit_rule, action):
        """
        # {u'bandwidth_limit_rule': {
        #              u'max_kbps': u'300',
        #              u'max_burst_kbps': u'30',
        #              'tenant_id': u'c414cf57e9fc49efb8e3e0c8d4e41b33',
        #              'id' :'aa1e84d8-9ea3-45ec-8197-83b7d9abc647'}}
        """
        LOG.info(_("rule_id = %(rule_id)s,policy_id = %(policy_id)s,"
                   "bandwith= %(bandwidth_limit_rule)s"),
                 {'rule_id': rule_id,
                  'policy_id': policy_id,
                  'bandwidth_limit_rule': bandwidth_limit_rule})
        if action == 'UPDATE':
            bandwidth_limit_rule = bandwidth_limit_rule['bandwidth_limit_rule']
        bandwidth_rule_info = {
            "input": {
                "id": rule_id if (rule_id != 0) else
                bandwidth_limit_rule['id'],
                "qos_policy_id": policy_id if (policy_id != 0)
                else bandwidth_limit_rule['policy_id'],
            }
        }
        input = bandwidth_rule_info['input']
        if 'max_kbps' in bandwidth_limit_rule:
            input['max_kbps'] = bandwidth_limit_rule['max_kbps']
        if 'max_burst_kbps' in bandwidth_limit_rule:
            input['max_burst_kbps'] = bandwidth_limit_rule['max_burst_kbps']
        if 'direction' in bandwidth_limit_rule:
            input['direction'] = bandwidth_limit_rule['direction']
        if 'tenant_id' in bandwidth_limit_rule:
            input['tenant_id'] = bandwidth_limit_rule['tenant_id']
        return bandwidth_rule_info

    @staticmethod
    def construct_dscp_marking_rule_info(context, rule_id, policy_id,
                                         dscp_marking_rule, action):
        LOG.info(_("rule_id = %(rule_id)s,policy_id = %(policy_id)s,"),
                 {'rule_id': rule_id,
                  'policy_id': policy_id,
                  'dscp_marking_rule': dscp_marking_rule})
        if action == 'UPDATE':
            dscp_marking_rule = dscp_marking_rule['dscp_marking_rule']
        dscp_rule_info = {
            "input": {
                "id": rule_id if (rule_id != 0) else
                dscp_marking_rule['id'],
                "qos_policy_id": policy_id if (policy_id != 0)
                else dscp_marking_rule['policy_id'],
            }
        }
        input = dscp_rule_info['input']
        if 'dscp_mark' in dscp_marking_rule:
            input['dscp_mark'] = dscp_marking_rule['dscp_mark']
        if 'tenant_id' in dscp_marking_rule:
            input['tenant_id'] = dscp_marking_rule['tenant_id']
        return dscp_rule_info

    @staticmethod
    def construct_delete_qos_info(id):
        qos_info = {
            "input": {
                "id": id
            }
        }
        return qos_info

    def create_policy(self, context, qos_policy):
        data = self.construct_qos_policy_info(context, 0, qos_policy,
                                              'CREATE')
        resource = 'add-' + QOS_POLICY
        err_str = _("Unable to add qos policy: %s")
        self.rest_action('POST', resource, data, err_str)

    def update_policy(self, context, id, qos_policy):
        data = self.construct_qos_policy_info(context, id, qos_policy,
                                              'UPDATE')
        resource = 'update-' + QOS_POLICY
        err_str = _("Unable to update qos policy: %s")
        self.rest_action('POST', resource, data, err_str)

    def delete_policy(self, context, id):
        data = self.construct_delete_qos_info(id)
        resource = 'del-' + QOS_POLICY
        err_str = _("Unable to add qos policy: %s")
        self.rest_action('POST', resource, data, err_str)

    def create_policy_bandwidth_limit_rule(self, context, policy_id,
                                           policy_rule):
        data = self.construct_bandwidth_limit_rule_info(context, 0, policy_id,
                                                        policy_rule, 'CREATE')
        resource = 'add-' + QOS_BANDWIDTH_RULE
        err_str = _("Unable to add bandwidth limit rule: %s")
        self.rest_action('POST', resource, data, err_str)

    def update_policy_bandwidth_limit_rule(self, context, rule_id, policy_id,
                                           policy_rule):
        data = self.construct_bandwidth_limit_rule_info(context, rule_id,
                                                        policy_id, policy_rule,
                                                        'UPDATE')
        resource = 'update-' + QOS_BANDWIDTH_RULE
        err_str = _("Unable to update bandwidth limit rule: %s")
        self.rest_action('POST', resource, data, err_str)

    def delete_policy_bandwidth_limit_rule(self, context, rule_id, policy_id):
        data = self.construct_delete_qos_info(rule_id)
        resource = 'del-' + QOS_BANDWIDTH_RULE
        err_str = _("Unable to del bandwidth limit rule: %s")
        self.rest_action('POST', resource, data, err_str)

    def create_policy_dscp_marking_rule(self, context, policy_id,
                                        policy_rule):
        data = self.construct_dscp_marking_rule_info(context, 0, policy_id,
                                                 policy_rule, 'CREATE')
        resource = 'add-' + QOS_DSCP_RULE
        err_str = _("Unable to add dscp marking rule: %s")
        self.rest_action('POST', resource, data, err_str)

    def update_policy_dscp_marking_rule(self, context, rule_id, policy_id,
                                    policy_rule):
        data = self.construct_dscp_marking_rule_info(context, rule_id,
                                                 policy_id, policy_rule,
                                                 'UPDATE')
        resource = 'update-' + QOS_DSCP_RULE
        err_str = _("Unable to update dscp marking rule: %s")
        self.rest_action('POST', resource, data, err_str)

    def delete_policy_dscp_marking_rule(self, context, rule_id, policy_id):
        data = self.construct_delete_qos_info(rule_id)
        resource = 'del-' + QOS_DSCP_RULE
        err_str = _("Unable to del dscp marking rule: %s")
        self.rest_action('POST', resource, data, err_str)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2017 ZTE, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_zte.common import servermanager
from neutron.db import common_db_mixin as base_db
from oslo_log import log

LOG = log.getLogger(__name__)

# The following are used to invoke the API on the external controller

FIREWALL = 'firewall'
FIREWALL_POLICY = 'firewall-policy'
FIREWALL_RULE = 'firewall-rule'
INSERT_RULE = 'insert-rule'
REMOVE_RULE = 'remove-rule'


BASE_URI = '/restconf/operations/zenic-fwapp-model:'
SUCCESS_CODES = range(200, 207)
FAILURE_CODES = [
    0,
    9,
    301,
    302,
    303,
    400,
    401,
    403,
    404,
    409,
    500,
    501,
    502,
    503,
    504,
    505]


class ZenicRestServerPool(
        servermanager.ServerPool,
        base_db.CommonDbMixin):
    """ZenicRestServerPool for firewall plugin.
    This server pool has firewall firewall_policy firewall_rule operations
    of create, update and delete, to the Zenic Controller.
    """

    def __init__(self, servers, auth, ssl, no_ssl_validation, ssl_sticky,
                 ssl_cert_directory, consistency_interval, timeout=False,
                 cache_connections=True, base_uri=BASE_URI,
                 success_codes=SUCCESS_CODES,
                 failure_codes=FAILURE_CODES, name='ZenicRestProxy'):
        super(ZenicRestServerPool, self).__init__(
            servers, auth, ssl, no_ssl_validation, ssl_sticky,
            ssl_cert_directory, consistency_interval, timeout,
            cache_connections, base_uri, success_codes, failure_codes, name)

    @staticmethod
    def construct_firewall_info(context, id, firewall, action):
        """
        :param context:
        :param id:
        :param firewall:
        :param action:
        :return: firewall={'status': 'INACTIVE',
           'router_ids': [],
           'name': u'fw2',
           'shared': None,
           'firewall_policy_id': u'58b61e09-b27a-439b-b428-aec51bf6061a',
           'tenant_id': u'a968288bfc9d4ced9c8c007f654310f9',
           'admin_state_up': True,
           'id': '3f2fb3ec-f7ce-450b-82e9-60983c10ff29',
           'description': u''}
        """
        if action == 'UPDATE':
            firewall = firewall['firewall']
        firewall_info = {
            "input": {
                "id": id if (id != 0) else firewall["id"]
            }
        }
        input = firewall_info['input']
        if 'status' in firewall:
            input['status'] = firewall['status']
        if 'router_ids' in firewall:
            input['router_ids'] = firewall['router_ids']
        if 'name' in firewall:
            input['name'] = firewall['name']
        if 'shared' in firewall:
            input['shared'] = firewall['shared']
        if 'tenant_id' in firewall:
            input['tenant_id'] = firewall['tenant_id']
        if 'firewall_policy_id' in firewall:
            input['firewall_policy_id'] = firewall['firewall_policy_id']
        if 'admin_state_up' in firewall:
            input['admin_state_up'] = firewall['admin_state_up']
        if 'description' in firewall:
            input['description'] = firewall['description']

        return firewall_info

    @staticmethod
    def construct_firewall_policy_info(context, id, firewall_policy, action):
        """
        :param context:
        :param id:
        :param firewall_policy:
        :param action:
        :return: firewall_policy={'name': u'fw5',
             'firewall_rules': [],
             'shared': False,
             'audited': False,
             'tenant_id': u'a968288bfc9d4ced9c8c007f654310f9',
             'id': '3c923e59-afed-4501-8e73-27be355746ab',
             'firewall_list': [],
             'description': u''}
        """
        if action == 'UPDATE':
            firewall_policy = firewall_policy['firewall_policy']
        firewall_policy_info = {
            "input": {
                "id": id if (id != 0) else firewall_policy['id'],
            }
        }
        input = firewall_policy_info['input']
        if 'firewall_rules' in firewall_policy:
            input['firewall_rules'] = firewall_policy['firewall_rules']
        if 'description' in firewall_policy:
            input['description'] = firewall_policy['description']
        if 'tenant_id ' in firewall_policy:
            input['tenant_id'] = firewall_policy['tenant_id']
        if 'shared' in firewall_policy:
            input['shared'] = firewall_policy['shared']
        if 'name' in firewall_policy:
            input['name'] = firewall_policy['name']
        if 'audited' in firewall_policy:
            input['audited'] = firewall_policy['audited']
        if 'firewall_list' in firewall_policy:
            input['firewall_list'] = firewall_policy['firewall_list']

        return firewall_policy_info

    @staticmethod
    def construct_firewall_rule_info(context, id, firewall_rule, action):
        """
        :param context:
        :param id:
        :param firewall_rule:
        :param action:
        :return:firewall_rule={'protocol': u'tcp',
            'description': u'',
            'source_port': None,
            'source_ip_address': None,
            'destination_ip_address': None,
            'firewall_policy_id': None,
            'position': 0,
            'destination_port': None,
            'id': '07410a40-3941-4e62-8f26-f87fc2032263',
            'name': u'fw5',
            'tenant_id': u'a968288bfc9d4ced9c8c007f654310f9',
            'enabled': True,
            'action': u'allow',
            'ip_version': 4,
            'shared': False}
        """
        if action == 'UPDATE':
            firewall_rule = firewall_rule['firewall_rule']
        firewall_rule_info = {
            "input": {
                "id": id if (id != 0) else firewall_rule['id'],
            }
        }
        input = firewall_rule_info['input']
        if 'protocol' in firewall_rule:
            input['protocol'] = firewall_rule['protocol']
        if 'description' in firewall_rule:
            input['description'] = firewall_rule['description']
        if firewall_rule.get('source_port', None):
            input['source_port'] = firewall_rule['source_port']
        if firewall_rule.get('source_ip_address', None):
            input['source_ip_address'] = firewall_rule['source_ip_address']
        if firewall_rule.get('destination_ip_address', None):
            input['destination_ip_address'] = firewall_rule[
                'destination_ip_address']
        if firewall_rule.get('destination_port', None):
            input['destination_port'] = firewall_rule['destination_port']
        if 'name' in firewall_rule:
            input['name'] = firewall_rule['name']
        if 'tenant_id' in firewall_rule:
            input['tenant_id'] = firewall_rule['tenant_id']
        if 'enabled' in firewall_rule:
            input['enabled'] = firewall_rule['enabled']
        if 'action' in firewall_rule:
            input['action'] = firewall_rule['action']
        if firewall_rule.get('firewall_policy_id', None):
            input['firewall_policy_id'] = firewall_rule['firewall_policy_id']
        if firewall_rule.get('position', 0):
            input['position'] = firewall_rule['position']
        if 'ip_version' in firewall_rule:
            input['ip_version'] = firewall_rule['ip_version']
        if 'shared' in firewall_rule:
            input['shared'] = firewall_rule['shared']

        return firewall_rule_info

    @staticmethod
    def construct_insert_rule_info(policy_id, insert_rule_info):
        """
        :param policy_id:
        :param insert_rule_info:
        :return: rule_info={u'insert_after': u'',
            u'firewall_rule_id': u'96541466-d916-43f2-926e-24c48939db5c',
            u'insert_before': u'790f6db4-7f63-4697-b037-32a37a08a69f'}
        """
        rule_info = {
            "input": {
                "policy_id": policy_id,
                "firewall_rule_id": insert_rule_info['firewall_rule_id'],
            }
        }
        input_str = rule_info['input']
        if insert_rule_info['insert_after']:
            input_str['insert_after'] = insert_rule_info['insert_after']
        if insert_rule_info['insert_before']:
            input_str['insert_before'] = insert_rule_info['insert_before']
        return rule_info

    @staticmethod
    def construct_delete_firewall_info(id):
        firewall_info = {
            "input": {
                "id": id
            }
        }
        return firewall_info

    @staticmethod
    def construct_remove_firewall_rule_info(policy_id, rule_id):
        firewall_info = {
            "input": {
                "policy_id": policy_id,
                "firewall_rule_id": rule_id['firewall_rule_id'],
            }
        }
        return firewall_info

    def create_firewall(self, context, firewall):
        data = self.construct_firewall_info(context, 0, firewall, 'CREATE')
        resource = 'add-' + FIREWALL
        err_str = _("Unable to add firewall: %s")
        self.rest_action('POST', resource, data, err_str)

    def update_firewall(self, context, id, firewall):
        data = self.construct_firewall_info(context, id, firewall, 'UPDATE')
        resource = 'update-' + FIREWALL
        err_str = _("Unable to update firewall: %s")
        self.rest_action('POST', resource, data, err_str)

    def delete_firewall(self, context, id):
        data = self.construct_delete_firewall_info(id)
        resource = 'del-' + FIREWALL
        err_str = _("Unable to add firewall: %s")
        self.rest_action('POST', resource, data, err_str)

    def create_firewall_policy(self, context, firewall_policy):
        data = self.construct_firewall_policy_info(
            context, 0, firewall_policy, 'CREATE')
        resource = 'add-' + FIREWALL_POLICY
        err_str = _("Unable to add firewall policy: %s")
        self.rest_action('POST', resource, data, err_str)

    def update_firewall_policy(self, context, id, firewall_policy):
        data = self.construct_firewall_policy_info(
            context, id, firewall_policy, 'UPDATE')
        resource = 'update-' + FIREWALL_POLICY
        err_str = _("Unable to update firewall policy: %s")
        self.rest_action('POST', resource, data, err_str)

    def delete_firewall_policy(self, context, id):
        data = self.construct_delete_firewall_info(id)
        resource = 'del-' + FIREWALL_POLICY
        err_str = _("Unable to del firewall policy: %s")
        self.rest_action('POST', resource, data, err_str)

    def create_firewall_rule(self, context, firewall_rule):
        data = self.construct_firewall_rule_info(
            context, 0, firewall_rule, 'CREATE')
        resource = 'add-' + FIREWALL_RULE
        err_str = _("Unable to add firewall rule: %s")
        self.rest_action('POST', resource, data, err_str)

    def update_firewall_rule(self, context, id, firewall_rule):
        data = self.construct_firewall_rule_info(
            context, id, firewall_rule, 'UPDATE')
        resource = 'update-' + FIREWALL_RULE
        err_str = _("Unable to update firewall rule: %s")
        self.rest_action('POST', resource, data, err_str)

    def delete_firewall_rule(self, context, id):
        data = self.construct_delete_firewall_info(id)
        resource = 'del-' + FIREWALL_RULE
        err_str = _("Unable to del firewall rule: %s")
        self.rest_action('POST', resource, data, err_str)

    def insert_rule(self, context, policy_id, rule_info):
        data = self.construct_insert_rule_info(policy_id, rule_info)
        # resource = FIREWALL_POLICY + '-' + INSERT_RULE
        resource = 'insert-' + FIREWALL_POLICY + '-rule'
        err_str = _("Unable to insert rule: %s")
        self.rest_action('POST', resource, data, err_str)

    def remove_rule(self, context, policy_id, rule_id):
        data = self.construct_remove_firewall_rule_info(policy_id, rule_id)
        # resource = FIREWALL_POLICY + '-' + REMOVE_RULE
        resource = 'remove-' + FIREWALL_POLICY + '-rule'
        err_str = _("Unable to remove rule: %s")
        self.rest_action('POST', resource, data, err_str)



from configobj import ConfigObj
import ConfigParser
import netaddr
import sdn_patch_api as api
import sys

# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2017 ZTE, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
 1. zenic plugin install include ml2 proxydriver install
    and zenic l3 agent install
 2. zenic plugin install shell be combine to one and support
    path/unpatch operation
"""

"""
1. copy proxydriver to /usr/lib/python2.7/site-packages/neutron/plugins,
   chmod +x;
2. modify /etc/neutron/plugin.ini, add proxydriver to mechanism_drivers,
   add[RESTPROXY]; servers&server_auth, values must be specified by users;
3. modify /usr/lib/python2.7/site-packages/neutron-2015.1.0-py2.7.egg-info
   /entry_points.txt, under [neutron.ml2.mechanism_drivers], add proxydriver=
   neutron.plugins.proxydriver.proxy_neutron_driver:ProxyMechanismDriver
4. restart neutron-server
"""

"""
1. copy all_l3_agent_scheduler.py to ....neutron/scheduler/. ;
2. copy zenic_agent.py to ....neutron/agent/. ;
3. copy neutron-zenic-agent to /usr/bin/. ;
4. copy neutron-zenic-agent.service to /usr/lib/systemd/system/. ;
5. systemctl enable neutron-zenic-agent.service;
6. systemctl restart neutron-zenic-agent.service;
7. systemctl status neutron-zenic-agent.service;  #check service status
8. openstack-config --set /etc/neutron/neutron.conf default router_scheduler_
   driver neutron.scheduler.all_l3_agents_scheduler.AllL3AgentsScheduler;
9. systemctl restart neutron-server.service;
10. systemctl status neutron-server.service;
"""

copy_files = {
    'zenic/all_l3_agents_scheduler.py': '/usr/lib/python2.7/site-'
    'packages/neutron/scheduler/',
    'zenic/zenic_agent.py': '/usr/lib/python2.7/site-packages/neutron/agent/',
    'zenic/neutron-zenic-agent': '/usr/bin/',
    'zenic/neutron-zenic-agent.service': '/usr/lib/systemd/system/',
    'zenic/proxy_server_manager.py': '/usr/lib/python2.7/site-packages'
    '/neutron/common/',
    'fw/zenic_fwaas_plugin.py': '/lib/python2.7/site-packages/neutron_fwaas'
    '/services/firewall/',
    'fw/zenic_firewall_pool.py': '/lib/python2.7/site-packages/neutron_fwaas'
    '/services/firewall/',
    'fw/zenic_server_manager.py': '/lib/python2.7/site-packages/neutron'
    '/common/',
    'ztetaas': '/lib/python2.7/site-packages/neutron_taas/services/taas/'
    'service_drivers',
    'qos/zenic_qos_plugin.py': '/lib/python2.7/site-packages/neutron/'
    'services/qos/',
    'qos/zenic_qos_pool.py': '/lib/python2.7/site-packages/neutron/'
    'services/qos/',
}

install_setservs = (
    'chmod 755 /usr/bin/neutron-zenic-agent',
    'systemctl disable neutron-l3-agent.service',
    'systemctl stop neutron-l3-agent.service',
)

enable_zenic_agent = (
    'systemctl enable neutron-zenic-agent.service',
)

install_chkservs = (
    'openstack-nova-api.service',
    'neutron-server.service',
    'neutron-zenic-agent.service',
)

install_chkcmccservs = (
    'neutron-server.service',
    'neutron-zenic-agent.service',
)

remove_setservs = (
    'systemctl disable neutron-zenic-agent.service',
    'systemctl stop neutron-zenic-agent.service',
    'systemctl enable neutron-l3-agent.service',
)

remove_zenic_servs = (
    'systemctl disable neutron-zenic-agent.service',
    'systemctl stop neutron-zenic-agent.service',
)

# remove_zenic_files = {
#     'zenic/neutron-zenic-agent': '/usr/bin/',
#     'zenic/neutron-zenic-agent.service': '/usr/lib/systemd/system/',
# }

remove_chkservs = (
    'neutron-l3-agent.service',
    'neutron-server.service',
)


def RAW_INPUT(str):
    return raw_input(str)


def recover_old_version():
    # /usr/lib/python2.7/site-packages/
    # neutron-2015.1.1-py2.7.egg-info/entry_points.txt
    full_args = "find /usr/lib -name entry_points.txt | grep neutron-"
    status, output = api.excute_command(full_args)

    full_args = "sed -i '/^proxydriver/d' %s" % output
    api.excute_command(full_args)

    full_args = "sed -i '/^zte_firewall/d' %s" % output
    api.excute_command(full_args)

    full_args = "sed -i '/^zte_qos/d' %s" % output
    api.excute_command(full_args)

    full_args = "sed -i '/^mechanism_drivers/s/,proxydriver//' " \
                "/etc/neutron/plugin.ini"
    api.excute_command(full_args)

    remove_git_code()

    api.remove_files(copy_files)


def recover_config(configs):

    # /etc/neutron/neutron.conf
    config_neutron = configs["config_neutron"]
    config_neutron['DEFAULT']['router_scheduler_driver'] = \
        'neutron.scheduler.l3_agent_scheduler.ChanceScheduler'
    #trunk
    try:
        del config_neutron['vlan_trunk']['trunk_drivers']
    except Exception as e:
        api.print_log("There is no key of %s in the neutron.conf" % e)
    config_neutron.write()

    # /etc/neutron/plugin.ini
    # taas
    config_plugin = configs["config_plugin"]
    taas_function = get_taas_para(config_plugin)
    api.save_log("get_taas_para taas_function = %s" % taas_function)
    if taas_function == 'tecs2.0':
        reset_taas_config("/etc/neutron/taas_plugin.conf")
    elif taas_function == 'tecs6.0':
        reset_taas_config("/etc/neutron/plugin.ini")
    elif taas_function == 'tecs3.0':
        reset_taas_config("/etc/neutron/plugin.ini")
    elif taas_function == 'openstack':
        reset_taas_config("/etc/neutron/taas_plugin.ini")
    else:
        api.save_log("recover config can not find taas function!")

    if check_op_version(config_plugin):
        op_version = config_plugin["RESTPROXY"]["op_version"]
    else:
        op_version = "False"
    if op_version != "bigcloud_K":
        # security_group_api
        config_nova = configs["config_nova"]
        config_nova['DEFAULT']['security_group_api'] = 'nova'
        config_nova.write()

    full_args = "sed -i '/^\[RESTPROXY/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '/^enable_qos/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '/^enable_M_qos/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '/^enable_bandwidth/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '/^enable_hierarchical_port/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '/^op_version/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '/^server_timeout/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '/^servers/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '/^server_auth/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '/^server_ssl/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '/^no_ssl_validation/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '/^enable_pre_commit/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '/^taas_function/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "grep ^mechanism_drivers[[:space:]]*=.*," \
                "[[:space:]]*zte_ml2driver /etc/neutron/plugin.ini"
    status, output = api.excute_command(full_args)
    if output != '':
        full_args = "sed -i '/^mechanism_drivers/s/," \
                    "[[:space:]]*zte_ml2driver//' /etc/neutron/plugin.ini"
        api.excute_command(full_args)
    full_args = "grep ^mechanism_drivers[[:space:]]*=.*zte_ml2driver" \
                "[[:space:]]*, /etc/neutron/plugin.ini"
    status, output = api.excute_command(full_args)
    if output != '':
        full_args = "sed -i '/^mechanism_drivers/s/zte_ml2driver" \
                    "[[:space:]]*,//' /etc/neutron/plugin.ini"
        api.excute_command(full_args)

    # remove l3 router_no_schedule
    full_args = "sed -i '/^\[L3\]/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '/^router_no_schedule/d' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "grep ^service_plugins[[:space:]]*=.*,[[:space:]]*zte_qos " \
                "/etc/neutron/neutron.conf"
    api.print_log("%s" % full_args)
    status, output = api.excute_command(full_args)
    if output != '':
        full_args = "sed -i '/^service_plugins/s/,[[:space:]]*zte_qos//' " \
                    "/etc/neutron/neutron.conf"
        api.print_log("%s" % full_args)
        api.excute_command(full_args)
        full_args = "sed -i '/^service_plugins/s/$/,qos/' " \
                    "/etc/neutron/neutron.conf"
        api.print_log("%s" % full_args)
        api.excute_command(full_args)

    full_args = "grep ^service_plugins[[:space:]]*=.*zte_qos[[:space:]]*, " \
                "/etc/neutron/neutron.conf"
    api.print_log("%s" % full_args)
    status, output = api.excute_command(full_args)
    if output != '':
        full_args = "sed -i '/^service_plugins/s/zte_qos[[:space:]]*,//' " \
                    "/etc/neutron/neutron.conf"
        api.print_log("%s" % full_args)
        api.excute_command(full_args)
        full_args = "sed -i '/^service_plugins/s/$/,qos/' " \
                    "/etc/neutron/neutron.conf"
        api.print_log("%s" % full_args)
        api.excute_command(full_args)


def set_taas_config(path):
    if api.check_if_file_exist(path):
        config_taas = ConfigObj(path, encoding='UTF8')
        if 'service_providers' not in config_taas.keys():
            config_taas['service_providers'] = {}
        config_taas['service_providers']['service_provider'] = \
            'TAAS:TAAS:networking_zte.taas.zte_taas_driver.ZTETaasDriver' \
            ':default'
        config_taas.write()
    return True


def delete_tecs2_taas_config(path):
    if api.check_if_file_exist(path):
        config_taas = ConfigObj(path, encoding='UTF8')
        if 'service_providers' in config_taas.keys():
            del config_taas['service_providers']
        config_taas.write()
    return True


def reset_taas_config(path):
    try:
        config_taas = ConfigObj(path, encoding='UTF8')
        config_taas['service_providers']['service_provider'] = \
            'TAAS:TAAS:neutron_taas.services.taas.service_drivers.' \
            'taas_rpc.TaasRpcDriver:default'
        config_taas.write()
    except Exception as e:
        api.print_log("no taas config found:%s,skip..." % e)


def remove_git_code():
    full_args = "rpm -qa|grep networking-zte|xargs rpm -e"
    api.print_log("%s" % full_args)
    api.excute_command(full_args)

    full_args = "rm -rf /usr/lib/python2.7/site-packages/networking_zte/"
    api.excute_command(full_args)


def append_config(kwargs, configs):
    config_neutron = ConfigObj('/etc/neutron/neutron.conf', encoding='UTF8')
    config_neutron['DEFAULT']['router_scheduler_driver'] = \
        'neutron.scheduler.l3_agent_scheduler.ChanceScheduler'

    # trunk
    if 'vlan_trunk' in config_neutron.keys():
        config_neutron['vlan_trunk']['trunk_drivers'] = 'networking_zte.' \
                                                        'trunk.driver'
    else:
        config_neutron['vlan_trunk'] = {}
        config_neutron['vlan_trunk']['trunk_drivers'] = 'networking_zte.' \
                                                        'trunk.driver'
    config_neutron.write()
    #taas
    if kwargs["taas_function"] == 'tecs2.0':
        if not set_taas_config("/etc/neutron/taas_plugin.conf"):
            return False
    elif kwargs["taas_function"] == 'tecs6.0':
        if not set_taas_config("/etc/neutron/plugin.ini"):
            return False
    elif kwargs["taas_function"] == 'tecs3.0':
        if not set_taas_config("/etc/neutron/plugin.ini"):
            return False
    elif kwargs["taas_function"] == 'openstack':
        if not api.check_if_file_exist('/etc/neutron/taas_plugin.ini'):
            api.print_log("use taas funtion of openstack while taas_plugin.ini"
                          " does not exist!")
            remove_git_code()
            return False
        if not set_taas_config("/etc/neutron/taas_plugin.ini"):
            return False
    else:
        api.save_log("append_config can not find taas function!")

    # plugin.ini
    full_args = "sed -i '$a \[RESTPROXY]' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '$a \enable_qos=%s' %s" % (
        kwargs["enable_qos"], "/etc/neutron/plugin.ini")
    api.excute_command(full_args)

    full_args = "sed -i '$a \enable_M_qos=%s' %s" % (
        kwargs["enable_M_qos"], "/etc/neutron/plugin.ini")
    api.excute_command(full_args)

    full_args = "sed -i '$a \op_version=%s' %s" % (
        kwargs["op_version"], "/etc/neutron/plugin.ini")
    api.excute_command(full_args)

    full_args = "sed -i '$a \enable_bandwidth=%s' %s" % (
        kwargs["enable_bandwidth"], "/etc/neutron/plugin.ini")
    if kwargs["enable_bandwidth"] is not None:
        api.excute_command(full_args)

    full_args = "sed -i '$a \enable_hierarchical_port=%s' %s" % (
        kwargs["enable_hierarchical_port"], "/etc/neutron/plugin.ini")
    if kwargs["enable_hierarchical_port"] is not None:
        api.excute_command(full_args)

    full_args = "sed -i '$a \server_timeout=100' %s" % (
        "/etc/neutron/plugin.ini")
    api.excute_command(full_args)

    full_args = "sed -i '$a \servers=%s' %s" % (
        kwargs["servers"], "/etc/neutron/plugin.ini")
    api.excute_command(full_args)

    full_args = "sed -i '$a \server_auth=%s' %s" % (
        kwargs["server_auth"], "/etc/neutron/plugin.ini")
    api.excute_command(full_args)

    full_args = "sed -i '$a \server_ssl=%s' %s" % (
        kwargs["ssl"], "/etc/neutron/plugin.ini")
    api.excute_command(full_args)

    full_args = "sed -i '$a \\no_ssl_validation=%s' %s" % (
        kwargs["ssl_validation"], "/etc/neutron/plugin.ini")
    api.excute_command(full_args)

    full_args = "sed -i '$a \\enable_pre_commit=%s' %s" % (
        kwargs["pre_commit"], "/etc/neutron/plugin.ini")
    api.excute_command(full_args)

    full_args = "sed -i '$a \\taas_function=%s' %s" % (
        kwargs["taas_function"], "/etc/neutron/plugin.ini")
    api.excute_command(full_args)

    full_args = "sed -i 's/^mechanism_drivers *=/mechanism_drivers " \
                "=zte_ml2driver,/g' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    # add l3 router_no_schedule
    full_args = "sed -i '$a \[L3\]' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    full_args = "sed -i '$a router_no_schedule=True' /etc/neutron/plugin.ini"
    api.excute_command(full_args)

    # full_args = "sed -i '/^service_plugins/s/$/,zte_firewall/'
    #/etc/neutron/neutron.conf"
    # api.excute_command(full_args)

    if kwargs["enable_M_qos"] == "True":
        api.save_log("using M qos!")
        full_args = "grep ^service_plugins[[:space:]]*=.*,[[:space:]]*qos " \
                    "/etc/neutron/neutron.conf"
        status, output = api.excute_command(full_args)
        if output != '':
            full_args = "sed -i '/^service_plugins/s/,[[:space:]]*qos//' " \
                        "/etc/neutron/neutron.conf"
            api.excute_command(full_args)
        full_args = "grep ^service_plugins[[:space:]]*=.*qos[[:space:]]*, " \
                    "/etc/neutron/neutron.conf"
        status, output = api.excute_command(full_args)
        if output != '':
            full_args = "sed -i '/^service_plugins/s/qos[[:space:]]*,//' " \
                        "/etc/neutron/neutron.conf"
            api.excute_command(full_args)

        full_args = "sed -i '/^service_plugins/s/$/,zte_qos/' " \
                    "/etc/neutron/neutron.conf"
        api.excute_command(full_args)

    if kwargs["op_version"] != "bigcloud_K":
        # security_group_api
        config_nova = configs["config_nova"]
        config_nova['DEFAULT']['security_group_api'] = 'neutron'
        config_nova.write()

    return True


def do_tecs_patch():
    try:
        import py_compile
        full_args = 'rm -f /usr/lib/python2.7/site-packages/neutron/plugins' \
                    '/proxydriver/common/rest/znic_l2/config.py'
        status, output = api.excute_command(full_args)
        api.print_log(output)
        full_args = 'cp -af config.py ' \
                    '/usr/lib/python2.7/site-packages/neutron/plugins' \
                    '/proxydriver/common/rest/znic_l2/config.py '
        status, output = api.excute_command(full_args)
        api.print_log(output)
        py_compile.compile('/usr/lib/python2.7/site-packages/neutron/plugins/'
                           'proxydriver/common/rest/znic_l2/config.py')
        return status
    except Exception as e:
        api.print_log("The reason for the failure of patch is %s!" % e)


def do_patch(op_patch):
    try:
        full_args = 'patch /usr/lib/python2.7/site-packages/neutron/' \
                    'services/l3_router/l3_router_plugin.py ' \
                    '<%s' % op_patch
        status, output = api.excute_command(full_args)
        api.print_log(output)
        return status
    except Exception as e:
        api.print_log("The reason for the failure of patch is %s!" % e)


def handle_auto_op_version(op_version):
    if not l3_scheduler_patch(op_version):
        return False
    return restart_neutron_server()


def handle_manual_op_version(op_version, configs):
    if not l3_scheduler_patch(op_version):
        return False
    full_args = "sed -i '/^mechanism_drivers/s/zte_ml2driver,//' " \
                "/etc/neutron/plugin.ini"
    api.excute_command(full_args)
    config = configs["config_plugin"]
    mechanism_drivers = config['ml2']['mechanism_drivers']
    api.print_log("patched mechanism_drivers is %s" % mechanism_drivers)
    return restart_neutron_server()


def l3_scheduler_patch(op_version):
    if op_version == 'tecs2.0' and api.check_if_file_exist(
            '/usr/lib/python2.7/site-packages/neutron/plugins/proxydriver'
            '/common/rest/znic_l2/config.py'):
        do_tecs_patch()
    value = 0
    if op_version in ['tecs2.0', 'tecs3.0', 'tecs6.0']:
        print ('This is %s, ignore patch '
              'to l3_router_plugin...\n' % op_version)
    else:
        if op_version in ['K', 'bigcloud_K']:
            value = do_patch('l3_scheduler_k.patch')
        elif op_version is 'L':
            value = do_patch('l3_scheduler_l.patch')
        elif op_version is 'M':
            value = do_patch('l3_scheduler_m.patch')
        elif op_version is 'N':
            value = do_patch('l3_scheduler_n.patch')
        elif op_version in ['O', 'bigcloud_O']:
            value = do_patch('l3_scheduler_o.patch')
    if value not in (0, 256, 1024):
        api.print_log("patch to l3_router_plugin failed!")
        return False

    if op_version == "bigcloud_K":
        install_openvswitch()

    return True


def restart_neutron_server():
    service = (
        'neutron-server.service',
    )
    return api.check_services(service)


def remove_plugin(configs):
    check_service_plugins("uninstall")
    api.excute_commads(remove_setservs)
    remove_openvswitch(configs)
    recover_old_version()
    recover_config(configs)
    remove_rpm_folder()
    return api.check_services(remove_chkservs)


def remove_rpm_folder():
    full_args = "rm -fr /usr/lib/python2.7/site-packages" \
                "/networking_zte-*.egg-info"
    api.print_log("%s" % full_args)
    api.excute_command(full_args)


def check_service_plugins(action):
    full_args = "grep ^service_plugins /etc/neutron/neutron.conf"
    api.print_log("%s" % full_args)
    status, output = api.excute_command(full_args)
    api.print_log("action:%s,service plugins:%s" % (action, output))
    full_args = "grep ^mechanism_drivers /etc/neutron/plugin.ini"
    api.print_log("%s" % full_args)
    status, output = api.excute_command(full_args)
    api.print_log("action:%s,mechanism_drivers:%s" % (action, output))


def patch_k_qos(qos_patch):
    full_args = 'patch /usr/lib/python2.7/site-packages/networking_zte/qos/' \
                'zenic_qos_plugin.py <%s' % qos_patch
    status, output = api.excute_command(full_args)
    api.print_log(output)
    return status


def post_install(op_version):
    api.excute_commads(install_setservs)
    if op_version == 'tecs2.0':
        patch_k_qos('qos.patch')
    if op_version == "bigcloud_K":
        return api.check_services(install_chkcmccservs)
    else:
        return api.check_services(install_chkservs)


def install_plugin(kwargs, configs):
    check_service_plugins("install")
    full_args = "rpm -ivh networking-zte-*"
    api.print_log("%s" % full_args)
    api.excute_command(full_args)
    #api.append_files(copy_files)
    ret = append_config(kwargs, configs)
    if not ret:
        return ret
    return post_install(kwargs["op_version"])


def update_plugin(kwargs, configs):
    """update code"""
    check_service_plugins("update")
    remove_rpm_folder()
    recover_old_version()
    recover_config(configs)
    full_args = "rpm -Uvh networking-zte-*"
    api.print_log("%s" % full_args)
    api.excute_command(full_args)
    #api.append_files(copy_files)
    ret = append_config(kwargs, configs)
    if not ret:
        return ret
    return post_install(kwargs["op_version"])


def if_update_config():
    operation = None
    while (not operation):
        operation = RAW_INPUT(
            "whether update configuration [y/n](default is n):\n")
        if str.lower(operation) == 'n':
            return False
        elif str.lower(operation) == 'y' or operation == '':
            return True
        else:
            operation = None
    return False


def manual_input_version():
    print ('****Please select openstack version****\n')
    print ('[1] Kilo\n')
    print ('[2] Liberty\n')
    print ('[3] Mitaka\n')
    print ('[4] Newton\n')
    print ('[5] Ocata\n')
    print ('[6] bigcloud_Kilo\n')
    print ('[7] bigcloud_Ocata\n')
    print ('[8] tecs2.0\n')
    print ('[9] tecs3.0\n')
    print ('[10] tecs6.0\n')
    all_version = {'1': 'K', '2': 'L', '3': 'M', '4': 'N', '5': 'O',
                   '6': 'bigcloud_K', '7': 'bigcloud_O', '8': 'tecs2.0',
                   '9': 'tecs3.0', '10': 'tecs6.0'}
    op_version_num = None
    while (1):
        op_version_num = RAW_INPUT("Press 1 or 2 or ... 10 to select:")
        if op_version_num not in ['1', '2', '3', '4', '5', '6', '7',
                                  '8', '9', '10']:
            print ("input parameter invaild, please input again:")
            continue
        else:
            break
    return all_version[op_version_num]


def interact_menu(op_version):
    all_input_para = {}

    ip = None
    while (not ip):
        ip = RAW_INPUT("Please input zenic IP Address:\n")
        try:
            netaddr.IPAddress(ip)
            if (4 != len(ip.strip().split('.'))):
                ip = None
        except BaseException:
            ip = None

    port = 0
    while (port < 1 or port > 65535):
        port = RAW_INPUT("Please input zenic Port(default: 8181):\n")
        port = '8181' if not port else port
        try:
            port = int(port)
        except BaseException:
            port = 0

    server = ''.join([ip, ':', str(port)])
    all_input_para["servers"] = server

    user = None
    while (not user):
        user = RAW_INPUT("Please input user name:\n")

    password = None
    while (not password):
        password = RAW_INPUT("Please input password:\n")

    auth = ''.join([user, ':', password])
    all_input_para["server_auth"] = auth

    ssl_flag = None
    ssl = None
    ssl_validation = None
    while (not ssl_flag):
        ssl_flag = RAW_INPUT(
            "Please confirm whether to enable HTTPS [y/n](default is n):\n")
        if str.lower(ssl_flag) == 'n' or ssl_flag == '':
            ssl = False
            ssl_validation = True
            break
        elif str.lower(ssl_flag) == 'y':
            ssl = True
            ssl_validation = True
        else:
            ssl_flag = None
    all_input_para["ssl"] = ssl
    all_input_para["ssl_validation"] = ssl_validation
    operation = None
    enable_M_qos = None
    enable_qos = None
    while (not operation):
        try:
            operation = RAW_INPUT(
                "Please confirm whether to enable qos function [y/n]:\n")
            if str.lower(operation) == 'y':
                if op_version in ['bigcloud_K', 'bigcloud_O']:
                    enable_qos = "True"
                    enable_M_qos = "False"
                else:
                    enable_qos = "False"
                    enable_M_qos = "True"
            elif str.lower(operation) == 'n' or operation == '':
                    enable_qos = "False"
                    enable_M_qos = "False"
            else:
                operation = None
        except Exception:
            operation = None
    all_input_para["enable_qos"] = enable_qos
    all_input_para["enable_M_qos"] = enable_M_qos
    function = None
    taas_function = None
    while (not function):
        try:
            function = RAW_INPUT('Please confirm whether to enable'
                                 ' taas function [y/n]:\n')
            if str.lower(function) == 'y':
                if op_version in ['tecs2.0', 'tecs3.0', 'tecs6.0']:
                    taas_function = op_version
                else:
                    taas_function = 'openstack'
            elif str.lower(function) == 'n' or function == '':
                taas_function = 'off'
            else:
                function = None
        except Exception:
            function = None
    all_input_para["taas_function"] = taas_function

    return all_input_para


def keep_zenic_agent():
    operation = None
    while (not operation):
        operation = RAW_INPUT(
            "whether zenic_agent keep alive [y/n](default is Y):\n")
        if str.lower(operation) == 'n':
            api.excute_commads(remove_zenic_servs)
            #api.remove_files(remove_zenic_files)
        elif str.lower(operation) == 'y' or operation == '':
            return True
        else:
            operation = None
    return True


def welcome():
    api.print_log("")
    api.print_log("================================================")
    api.print_log("    ZTE SDN Plugin Installation Wizard")
    api.print_log("================================================")


def menu():
    api.print_log("")
    api.print_log("1. install")
    api.print_log("2. uninstall")
    api.print_log("3. update(eg: tecs2.0->tecs2.0)")
    api.print_log("4. update_tecs6.0(tecs2.0->tecs6.0)")
    api.print_log("5. exit")
    api.print_log("")


def operation_menu():
    menu()
    operation = None
    while (not operation):
        try:
            operation = RAW_INPUT("Please select an operation: ")
            if operation == '1':
                action = "install"
            elif operation == '2':
                action = "uninstall"
            elif operation == '3':
                action = "update"
            elif operation == '4':
                action = "update_tecs6.0"
            elif operation == '5':
                action = "exit"
            else:
                operation = None
        except Exception:
            operation = None
    return action


def check_zenic_para(configs):
    config = configs["config_plugin"]
    para = config.keys()
    if 'RESTPROXY' in para:
        return True
    else:
        return False


def install_openvswitch():
    if api.check_if_file_exist('/usr/lib/python2.7'
                               '/site-packages/neutron/plugins/ml2/drivers/'
                               'mech_agent.py'):
        full_args = 'mv /usr/lib/python2.7/site-packages/neutron/' \
                 'plugins/ml2/drivers/mech_agent.py /usr/lib/python2.7/' \
                 'site-packages/neutron/plugins/ml2/drivers/mech_agent.py_ovs'
        status, output = api.excute_command(full_args)
    update_openvswitch()


def update_openvswitch():
    full_args = 'rm -fr /usr/lib/python2.7/site-packages/neutron/plugins'\
        '/ml2/drivers/mech_agent.py'
    status, output = api.excute_command(full_args)
    full_args = 'cp mech_agent.py /usr/lib/python2.7/site-packages'\
        '/neutron/plugins/ml2/drivers/'
    status, output = api.excute_command(full_args)


def remove_openvswitch(configs):
    config = configs["config_plugin"]
    if check_op_version(config):
        op_version = config["RESTPROXY"]["op_version"]
        if op_version == "bigcloud_K":
            full_args = 'rm -fr /usr/lib/python2.7/site-packages/neutron'\
                '/plugins/ml2/drivers/mech_agent.py'
            status, output = api.excute_command(full_args)
            full_args = 'mv /usr/lib/python2.7/site-packages/neutron/plugins'\
                '/ml2/drivers/mech_agent.py_ovs /usr/lib/python2.7/'\
                'site-packages/neutron/plugins/ml2/drivers/mech_agent.py'
            status, output = api.excute_command(full_args)


def check_qos_para(config):
    if 'RESTPROXY' in config.keys():
        zenic_para = config['RESTPROXY'].keys()
        if 'enable_qos' in zenic_para:
            return True
        else:
            return False
    else:
        return False


def check_op_version(config):
    if 'RESTPROXY' in config.keys():
        zenic_para = config['RESTPROXY'].keys()
        if 'op_version' in zenic_para:
            return True
        else:
            return False
    else:
        return False


def check_m_qos_para(config):
    if 'RESTPROXY' in config.keys():
        zenic_para = config['RESTPROXY'].keys()
        if 'enable_M_qos' in zenic_para:
            return True
        else:
            return False
    else:
        return False


def check_bandwidth_para(config):
    if 'RESTPROXY' in config.keys():
        zenic_para = config['RESTPROXY'].keys()
        if 'enable_bandwidth' in zenic_para:
            return True
        else:
            return False
    else:
        return False


def check_hierarchical_port_para(config):
    if 'RESTPROXY' in config.keys():
        zenic_para = config['RESTPROXY'].keys()
        if 'enable_hierarchical_port' in zenic_para:
            return True
        else:
            return False
    else:
        return False


def check_ssl_para(config):
    if 'RESTPROXY' in config.keys():
        zenic_para = config['RESTPROXY'].keys()
        if 'server_ssl' in zenic_para and 'no_ssl_validation' in zenic_para:
            return True
        else:
            return False
    else:
        return False


def check_pre_commit_para(config):
    if 'RESTPROXY' in config.keys():
        zenic_para = config['RESTPROXY'].keys()
        if 'enable_pre_commit' in zenic_para:
            return True
        else:
            return False
    else:
        return False


def check_taas_para(config):
    if 'RESTPROXY' in config.keys():
        zenic_para = config['RESTPROXY'].keys()
        if 'taas_function' in zenic_para:
            return True
        else:
            return False
    else:
        return False


def get_taas_para(config):
    taas_function = 'off'
    if check_taas_para(config):
        taas_function = config["RESTPROXY"]["taas_function"]
        return taas_function

    # full_args = 'openstack-config --get /etc/neutron/plugin.ini '\
    #             'service_providers service_provider'
    # status, output = api.excute_command(full_args)
    # if not status:
    #     taas_function = 'tecs3.0'
    #     return taas_function
    #
    config_tass = ConfigObj('/etc/neutron/taas_plugin.conf')
    if 'service_providers' in config_tass.keys():
        if 'service_provider' in config_tass['service_providers'].keys():
            taas_function = 'tecs2.0'
    return taas_function


def get_zenic_para(configs):
    all_input_para = {}
    config = configs["config_plugin"]
    all_input_para["servers"] = config['RESTPROXY']['servers']
    ini_config = ConfigParser.ConfigParser()
    ini_config.readfp(open('/etc/neutron/plugin.ini', "r"))
    all_input_para["server_auth"] = ini_config.get("RESTPROXY", "server_auth")
    if check_qos_para(config):
        all_input_para["enable_qos"] = config['RESTPROXY']['enable_qos']
    else:
        all_input_para["enable_qos"] = "False"
    if check_bandwidth_para(config):
        all_input_para["enable_bandwidth"] = \
            config['RESTPROXY']['enable_bandwidth']
    else:
        all_input_para["enable_bandwidth"] = None
    if check_hierarchical_port_para(config):
        all_input_para["enable_hierarchical_port"] = \
            config['RESTPROXY']['enable_hierarchical_port']
    else:
        all_input_para["enable_hierarchical_port"] = None
    if check_ssl_para(config):
        all_input_para["ssl"] = config['RESTPROXY']['server_ssl']
        all_input_para["ssl_validation"] = \
            config['RESTPROXY']['no_ssl_validation']
    else:
        all_input_para["ssl"] = False
        all_input_para["ssl_validation"] = True

    if check_pre_commit_para(config):
        all_input_para["pre_commit"] = \
            config['RESTPROXY']['enable_pre_commit']
    else:
        all_input_para["pre_commit"] = "False"
    if check_m_qos_para(config):
        all_input_para["enable_M_qos"] = \
            config['RESTPROXY']['enable_M_qos']
    else:
        all_input_para["enable_M_qos"] = "False"
    all_input_para["taas_function"] = get_taas_para(config)
    api.save_log("get_taas_para taas_function = %s" %
                 all_input_para["taas_function"])

    if check_op_version(config):
        all_input_para["op_version"] = config['RESTPROXY']['op_version']
    else:
        all_input_para["op_version"] = None

    return all_input_para


def check_ml2driver_if_install(configs):
    config = configs["config_plugin"]
    drivers = config['ml2']['mechanism_drivers']
    if 'zte_ml2driver' in drivers or 'proxydriver' in drivers:
        return True
    if api.check_if_file_exist('/usr/lib/python2.7'
                               '/site-packages/networking_zte'):
        return True
    full_args = "ls /usr/lib/python2.7/site-packages/ | grep networking_zte-"
    status, output = api.excute_command(full_args)
    if "egg-info" in output:
        return True
    return False


def get_all_para(configs, op_version):
    if check_zenic_para(configs):
        all_input_para = get_zenic_para(configs)

    else:
        all_input_para = interact_menu(op_version)
        all_input_para["pre_commit"] = "False"
        all_input_para["enable_bandwidth"] = None
        all_input_para["enable_hierarchical_port"] = None

    all_input_para["op_version"] = op_version
    return all_input_para


def update_tecs_six(all_input_para):
    if api.check_if_file_exist('/usr/lib/python2.7/site-packages/neutron/'
                               'plugins/proxydriver/common/rest/znic_l2'
                               '/config.py'):
        do_tecs_patch()
    if all_input_para["op_version"] is None or all_input_para["op_version"] \
            == 'tecs2.0':
        all_input_para["op_version"] = 'tecs6.0'
    if all_input_para["taas_function"] == 'tecs2.0':
        all_input_para["taas_function"] = 'tecs6.0'
    if api.check_if_file_exist('/etc/neutron/taas_plugin.conf'):
        delete_tecs2_taas_config('/etc/neutron/taas_plugin.conf')
    return all_input_para


def refact_params(input_dict):
    # change old config version to new config version
    if 'taas_function' not in input_dict \
            or input_dict['taas_function'] == 'openstack':
        input_dict['taas_function'] = 'off'
    return input_dict


def zenic_plugin_auto_install(auto_install, configs):
    action = auto_install
    if action == 'uninstall':
        if not remove_plugin(configs):
            return 1
        else:
            return 0

    # auto install must have plugin config in plugin.ini
    if not check_zenic_para(configs):
        api.print_log("Can not find sdn plugin config in plugin.ini!")
        return 1
    # get op_version and do patch files and restart neutron-server first
    config_plugin = configs["config_plugin"]
    if check_op_version(config_plugin):
        op_version = config_plugin["RESTPROXY"]["op_version"]
        if not handle_auto_op_version(op_version):
            return 1
    else:
        api.print_log("Sorry this plugin version does not support auto "
                      "install, please use manual install instead!")
        return 1
    # auto-install only deal with rpm package
    check_service_plugins(action)
    remove_rpm_folder()
    remove_git_code()
    if action == 'install':
        full_args = "rpm -ivh networking-zte-*"
    if action == 'update':
        full_args = "rpm -Uvh networking-zte-*"
    api.print_log("%s" % full_args)
    api.excute_command(full_args)

    if not post_install(op_version):
        api.print_log("\33[30mzenic plugin %s failed, please "
                      "fix and try again!!\33[0m" % action)
        return 1
    else:
        api.print_log("\33[36mzenic plugin %s success!\33[0m" % action)
        return 0


def zenic_plugin_install(configs):
    welcome()
    op_version = manual_input_version()
    if not handle_manual_op_version(op_version, configs):
        return 1
    action = operation_menu()
    api.print_log("User select to %s plugin...\n" % action)
    if action == 'install' and check_ml2driver_if_install(configs):
        api.print_log("zenic plugin has already installed!")
        return 1
    if action == 'update' and not check_ml2driver_if_install(configs):
        api.print_log("zenic plugin has not installed! "
                      "please install first!")
        return 1

    if action == 'install':
        all_input_para = get_all_para(configs, op_version)

        api.print_log("Start to %s plugin...\n" % action)

        rtn = install_plugin(all_input_para, configs)
        if rtn:
            keep_zenic_agent()
        api.excute_commads(enable_zenic_agent)
    elif action == 'update' or action == 'update_tecs6.0':
        all_input_para = get_all_para(configs, op_version)

        all_input_para = refact_params(all_input_para)

        if action == 'update_tecs6.0':
            all_input_para = update_tecs_six(all_input_para)
        api.print_log("Start to %s plugin...\n" % action)
        rtn = update_plugin(all_input_para, configs)
        if rtn:
            keep_zenic_agent()
        api.excute_commads(enable_zenic_agent)
    elif action == 'uninstall':
        api.print_log("Start to %s plugin...\n" % action)
        rtn = remove_plugin(configs)
    else:
        rtn = True

    if rtn:
        api.print_log("\33[36mzenic plugin %s success!\33[0m" % action)
        return 0
    else:
        api.print_log("\33[30mzenic plugin %s failed, please "
                      "fix and try again!!\33[0m" % action)
        return 1


def check_environment_config():
    configs = {}
    try:
        config_neutron = ConfigObj('/etc/neutron/neutron.conf',
                                   encoding='UTF8')
    except Exception as e:
        api.print_log("There some errors in the neutron.conf %s!" % e)
        return configs
    try:
        config_plugin = ConfigObj('/etc/neutron/plugin.ini',
                                  encoding='UTF8')
    except Exception as e:
        api.print_log("There some errors in the plugin.ini %s!" % e)
        return configs
    try:
        config_nova = ConfigObj('/etc/nova/nova.conf',
                                encoding='UTF8')
    except Exception as e:
        api.print_log("There some errors in the nova.conf %s!" % e)
        return configs
    configs["config_neutron"] = config_neutron
    configs["config_plugin"] = config_plugin
    configs["config_nova"] = config_nova

    return configs


if __name__ == '__main__':
    configs = check_environment_config()
    if len(configs) != 0:
        api.print_log("------------------------------------------------------")
        version = api.get_version_info('VERSION', 'version')
        date = api.get_version_info('VERSION', 'date')
        api.print_log("version: %s = %s" % ('zenic_plugin_version', "%s" %
                                            date))
        api.backup_old_files()
        install_type = None
        if len(sys.argv) > 1:
            install_type = sys.argv[1]
            api.print_log("\nRecognizing auto: {} zenic plugin...\n".
                          format(install_type))
            rtn = zenic_plugin_auto_install(install_type, configs)
        else:
            rtn = zenic_plugin_install(configs)
        api.set_sys_info('VERSION', 'zenic_version', "kilo_v2.1.%s" % version)
        api.set_sys_info('VERSION', 'date', date)
        sys.exit(rtn)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2017 ZTE, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""""
 zte sdn patch install api
"""

import commands      # noqa
import ConfigParser
import logging
import os
import time

log_file = "install.log"
back_file = "patch.backup"
save_path = "/home/sdn_patch"
version_file = "version.info"
plugin_name = "networking_zte"
code_folder = "/usr/lib/python2.7/site-packages/%s/" % plugin_name
plugin_conf = "plugin.ini"
neutorn_conf = "neutron.conf"
taas_conf = "taas_plugin.conf"
if not os.path.exists(save_path):
    os.mkdir(save_path)


def mock_action():
    print_log("doing nothing, just for unittest")


def save_log(info):
    global save_path
    save_file = os.path.join(save_path, log_file)
    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s %(name)-2s %(levelname)-2s %(message)s',
        datefmt='%m-%d %H:%M',
        filename=save_file,
        filemode='a')
    logger = logging.getLogger()
    logger.info(info)


def print_log(info):
    print (info)
    save_log(info)


def excute_command(full_args):
    try:
        status, output = commands.getstatusoutput(full_args)
        save_log('cmd:%s' % full_args)
        save_log('out:[%d]%s' % (status, output))
        if status not in (0, 256, 1024):
            print_log('cmd fail:%s' % full_args)
        return (status, output)
    except Exception as e:
        print_log('cmd err:%s' % full_args)
        msg = _('excute_command: full_args=%(full_args)s failed, '
                'error: e=%(e)s.'), {'full_args': full_args, 'e': e}
        raise ValueError(msg)


def backup_old_files():
    print_log("bakup old version files:...")
    bak_path = os.path.join(save_path, 'bak')
    if os.path.exists(bak_path):
        full_args = "rm -rf %s" % bak_path
        excute_command(full_args)
    os.mkdir(bak_path)
    full_args = "cp -af %s %s/%s" % (code_folder, bak_path, plugin_name)
    print_log(full_args)
    excute_command(full_args)
    full_args = "cp -af /etc/neutron/%s %s/%s" % (plugin_conf, bak_path,
                                                  plugin_conf)
    print_log(full_args)
    excute_command(full_args)
    full_args = "cp -af /etc/neutron/%s %s/%s" % (neutorn_conf, bak_path,
                                                  neutorn_conf)
    print_log(full_args)
    excute_command(full_args)
    full_args = "cp -af /etc/neutron/%s %s/%s" % (taas_conf, bak_path,
                                                  taas_conf)
    print_log(full_args)
    excute_command(full_args)


def get_sys_info(section, option):
    global save_path
    fpath = os.path.join(save_path, back_file)
    conf = ConfigParser.ConfigParser()
    conf.read(fpath)
    try:
        value = conf.get(section, option)
    except BaseException:
        value = None

    print_log("read conf: [%s] %s = %s" % (section, option, value))
    return value


def get_version_info(section, option):
    conf = ConfigParser.ConfigParser()
    conf.read(version_file)
    try:
        value = conf.get(section, option)
    except BaseException:
        value = None

    save_log("read conf: [%s] %s = %s" % (section, option, value))
    return value


def set_sys_info(section, option, value):
    global save_path
    fpath = os.path.join(save_path, back_file)
    conf = ConfigParser.ConfigParser()
    conf.read(fpath)
    if not conf.has_section(section):
        conf.add_section(section)
    try:
        conf.set(section, option, value)
    except BaseException:
        save_log("save err: [%s] %s = %s" % (section, option, value))

    conf.write(open(fpath, 'w'))
    save_log("save conf: [%s] %s = %s" % (section, option, value))


def excute_commads(cmdslist):
    for cmd in cmdslist:
        print_log(cmd)
        excute_command(cmd)


def remove_files(filelist):
    for file_name in filelist:
        file_path = os.path.join(
            filelist[file_name],
            os.path.basename(file_name))
        if os.path.isfile(file_path):
            full_args = "rm -f %s" % file_path
        elif os.path.isdir(file_path):
            full_args = "rm -rf %s" % file_path
        else:
            print_log("rm miss: %s" % file_path)
            continue  # you must not return here, one error not all error
        print_log(full_args)
        excute_command(full_args)


def append_files(filelist):
    remove_files(filelist)
    for file_name in filelist:
        file_path = os.path.join(os.getcwd(), file_name)
        if os.path.isfile(file_path):
            full_args = "cp -fp %s %s" % (file_path, filelist[file_name])
        elif os.path.isdir(file_path):
            full_args = "cp -rfp %s %s" % (file_path, filelist[file_name])
        else:
            print_log("cp fail: %s %s" % (file_path, filelist[file_name]))
            continue  # you must not return here, one error not all error
        print_log(full_args)
        excute_command(full_args)


def restart_service_and_check(serv_name):
    full_args = 'systemctl restart %s' % serv_name
    print_log(full_args)
    excute_command(full_args)
    time.sleep(3)
    return check_service_only(serv_name)


def check_service_only(serv_name):
    full_args = 'systemctl status %s' % serv_name
    print_log(full_args)
    status, output = excute_command(full_args)
    if 'active (running)' not in output:
        print_log('Restarting %s fail.' % serv_name)
        return False
    return True


def restart_openstack_and_check():
    full_args = 'openstack-service restart'
    print_log(full_args)
    excute_command(full_args)
    time.sleep(5)
    full_args = 'openstack-service status'
    print_log(full_args)
    status, output = excute_command(full_args)
    if 'inactive' in output:
        print_log('Restarting openstack fail.')
        return False
    if 'failed' in output:
        print_log('Restarting openstack fail.')
        return False
    return True


'''
flag default value is false,check and restart service
if flag value is true only check service
'''


def check_services(servslist, flag=False):
    checkerror = 0
    for serv in servslist:
        print_log("checking service: %s" % serv)
        if flag:
            if not check_service_only(serv):
                checkerror += 1
        else:
            if not restart_service_and_check(serv):
                checkerror += 1
    if checkerror != 0:
        return False
    else:
        return True


def check_if_file_exist(filename):
    if not os.path.exists(filename):
        return False
    return True
